/**************************************************************************
 *   Inferno : An automatic semantic information extractor                 *
 *   Copyright (C) 2006-2010  University of Michigan                       *
 *   http://www.eecs.umich.edu/inferno                                     *
 *                                                                         *
 *   Contributors include Andrew DeOrio, Adam Bauserman,                   *
 *   Valeria Bertacco and Beth Isaksen.                                    *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA         *
 *   02110-1301  USA                                                       *
 *                                                                         *
 *   Andrew DeOrio <awdeorio@umich.edu>                                    *
 *   Valeria Bertacco <valeria@umich.edu>                                  *
 *                                                                         *
 *   University of Michigan                                                *
 *   Electrical Engineering and Computer Science Dept.                     *
 *   2260 Hayward St.                                                      *
 *   Ann Arbor, MI 48109-2121                                              *
 **************************************************************************/

#include <cstdlib>

#include "Config_Parser.h"

Config_Parser::Config_Parser(const string& config_file_name_in) : config_file_name(config_file_name_in) {}
Config_Parser::~Config_Parser() {}

bool Config_Parser::parse(string& output_file_base, string& top_name, int& max_lines, string& reset_name,
                          string& clock_name, bool& transaction_mode, bool& sync_mode, bool& signal_mode,
                          bool& all_signals, bool& start_mode, vector<string>& start_signal_names,
                          vector<string>& start_triggers, vector<string>& verilog_f_names, 
                          string& trace_file_name, int& max_bits, string& main_module,
                          set<string>& signals_to_analyze, set<string>& modules_to_analyze,
                          bool& loopback_edges, vector<string>& ignore_states, bool& show_hex,
                          bool& strip_leading_zeros) {
  //First set some defaults
  signal_mode = false;
  all_signals = false;
  transaction_mode = true;
  sync_mode = true;
  start_mode = false;
  max_lines = 100000;
  max_bits = 1;
  loopback_edges = false;
  ignore_states = vector<string>();
  show_hex = false;
  strip_leading_zeros = false;
        
        
  ifstream config_file(config_file_name.c_str());
  if(!config_file.is_open()) {
    mout << "Failed to open config file: " << config_file_name << endl;
    print_help();
    return 0;
  }
  else {
    mout << "Opened config file: " << config_file_name << endl;
  }

  bool got_output_file_base = false, got_top_name = false, got_max_lines = false,
    got_clock_name = false, got_reset_name = false, got_sync_mode = false,
    got_max_bits = false, got_modules = false, got_signals = false,
    got_transaction_mode = false, got_start_transaction = false,
    got_verilog_file_names = false, got_trace_file_name = false;
        
  vector<string> words;
  while(getWords(config_file,words,false," ")) {
    if(words.size() < 1) {
      print_help();
      return false;
    }
    if(words[0].at(0) == '#')
      {
        // ignore this line, it is a comment
      }
    else if(words[0] == "-name") {
      if(words.size() != 2) {
        mout << "Must provide exactly one name (file name base).\n";
        print_help();
        return false;
      }
      // append date/time string to output file base
      time_t tm = time(NULL);
      char time_str[BUFSIZ];
      strftime(time_str ,BUFSIZ , "-%Y-%m-%d_%H%M",localtime(&tm));
      output_file_base = words[1] + string(time_str);

      // make output/ directory for new folder to go in
      system("mkdir output 2> /dev/null");
      // make new folder
      system((string("mkdir ") + string("output/") + 
              output_file_base + string(" 2> /dev/null")
              ).c_str());
      output_file_base = "output/" + output_file_base + "/";
      got_output_file_base = true;
    }
    else if(words[0] == "-top") {
      if(words.size() != 2) {
        mout << "Must provide exactly one top module name.\n";
        print_help();
        return false;
      }
      top_name = words[1];
      got_top_name = true;
    }
    else if(words[0] == "-max_lines") {
      if(words.size() != 2) {
        mout << "Must provide exactly one maximum number of lines.\n";
        print_help();
        return false;
      }
      max_lines = atoi(words[1]);
      if(max_lines < 1) {
        mout << "max_lines must specify a positive integer." << endl;
        print_help();
        return false;
      }
      got_max_lines = true;
    }
    else if(words[0] == "-clock") {
      if(words.size() != 2) {
        mout << "Must provide exactly one clock name (sorry!).\n";
        print_help();
        return false;
      }
      clock_name = words[1];
      got_clock_name = true;
    }
    else if(words[0] == "-reset") {
      if(words.size() != 2) {
        mout << "Must provide exactly one reset name.\n";
        print_help();
        return false;
      }
      reset_name = words[1];
      got_reset_name = true;
    }
    else if(words[0] == "-synchronous") {
      sync_mode = true; //note: this is the default
      got_sync_mode = true;
    }
    else if(words[0] == "-asynchronous") {
      sync_mode = false;
      got_sync_mode = true;
    }
    else if(words[0] == "-max_bits") {
      if(words.size() != 2) {
        mout << "Must provide exactly one number of maximum bits for a signal to be considered.\n";
        print_help();
        return false;
      }
      max_bits = atoi(words[1]);
      if(max_bits < 1) {
        mout << "max_bits must specify a positive integer." << endl;
        print_help();
        return false;
      }
      got_max_bits = true;
    }
    else if(words[0] == "-modules") {
      if(words.size() < 2) {
        mout << "Must provide at least one module name for analysis.\n";
        print_help();
        return false;
      }
      for(unsigned int i = 1; i < words.size(); ++i) {
        modules_to_analyze.insert(words[i]);
        if(i == 1)
          main_module = words[i];
      }
      got_modules = true;
    }
    else if(words[0] == "-signals") {
      fout << "signal mode\n";
      signal_mode = true;
      if(words.size() < 2) {
        mout << "Must provide at least one signal to analyze.\n";
        exit(1);
      }
      if(words[1] == "-all") {
        all_signals = true;
      }
      else {
        for(unsigned int i = 1; i < words.size(); ++i) {
          fout << "Will analyze signal " << words[i] << endl;
          signals_to_analyze.insert(words[i]);
        }
      }
      got_signals = true;
    }
    else if(words[0] == "-ignore_states") {
      if(words.size() < 2) {
        mout << "Must provide at least one state to be ignored.\n";
        exit(1);
      }
      else {
        for(unsigned int i = 1; i < words.size(); ++i) {
          fout << "Will ignore state " << words[i] << endl;
          ignore_states.push_back(words[i]);
        }
      }
    }
    else if(words[0] == "-transaction_mode") {
      if(words.size() != 2) {
        mout << "Must say either -true or -false for transaction mode (default true).\n";
        print_help();
        return false;
      }
      if(words[1] == "-false") 
        transaction_mode = false;
      else if(words[1] == "-true")
        transaction_mode = true;
      else {
        mout << "Only options for transaction mode are -true and -false.\n";
        print_help();
        return false;
      }
      got_transaction_mode = true;
    }
    else if(words[0] == "-start_transaction") {
      if(words.size() != 3 || (words[2] != "01" && words[2] != "10")) {
        mout << "The -start_transaction option must be followed by a signal name and then either 01 (rising) or 10 (falling).\n";
        print_help();
        return false;
      }
      start_mode = true;
      start_signal_names.push_back(words[1]);
      if(words[2] == "01")
        start_triggers.push_back("rise");
      else
        start_triggers.push_back("fall");
      got_start_transaction = true;
    }
    else if(words[0] == "-verilog") {
      if(words.size() < 2) {
        mout << "Must provide at least one verilog file (containing the description of the module under analysis).\n";
        print_help();
        return false;
      }
      for(unsigned int i = 1; i < words.size(); ++i) {
        verilog_f_names.push_back(words[i]);
        fout << "Adding verilog file name: " << words[i] << endl;
      }
      got_verilog_file_names = true;
    }
    else if(words[0] == "-vcd") {
      if(words.size() != 2) {
        mout << "Must provide exactly one vcd file (sorry!).\n";
        print_help();
        return false;
      }
      trace_file_name = words[1];
      got_trace_file_name = true;
    }
    else if(words[0] == "-loopback_edges") {
      if(words.size() != 1) {
        mout << "The -loopback_edges option should not be followed by any parameter(s).\n";
        print_help();
        return false;
      }
      loopback_edges = true;
    } 
    else if(words[0] == "-show_hex") {
      if(words.size() != 1) {
        mout << "The -show_hex option should not be followed by any parameter(s).\n";
        print_help();
        return false;
      }
      show_hex = true;
    } 
    else if(words[0] == "-strip_leading_zeros") {
      if(words.size() != 1) {
        mout << "The -strip_leading_zeros option should not be followed by any parameter(s).\n";
        print_help();
        return false;
      }
      strip_leading_zeros = true;
    } 
    else {
      mout << "Error: unexpected option: " << words[0] << endl;
      print_help();
      return false;
    }
  }
  //More checking -- to see if all necessary information was provided
  //and to inform the user in cases where the default will be used.
  if(!got_output_file_base) {
    mout << "A -name line must be included to specify the working "
         << "directory and base for the output file names.\n";
    print_help();
    return false;
  }
  else {
    fout << "Working directory/output file base is: " << output_file_base << endl;
  }
  if(!got_top_name) {
    mout << "A -top line must be included to specify the name of the top "
         << "level module." << endl;
    print_help();
    return false;
  }
  else {
    fout << "Top module name is: " << top_name << endl;
  }
  if(!got_max_lines) {
    mout << "No max_lines specified, using default value of " << max_lines << "."
         << endl;
  }
  else {
    fout << "Maximum number of lines to read from trace file is: " << max_lines << endl;
  }
  if(!got_clock_name) {
    if(sync_mode) {
      mout << "To use the synchronous mode a -clock line must be included "
           << "to specify the name of the clock signal." << endl;
      print_help();
      return false;
    }
    else {
      mout << "Warning, no clock signal specified." << endl;
    }
  }
  else {
    fout << "Name of clock signal is: " << clock_name << endl;
  }
  if(!got_reset_name) {
    //??? Should this be an error (sometimes)?
    mout << "Warning no reset specified." << endl;
  }
  else {
    fout << "Name of reset signal is: " << reset_name << endl;
  }
  if(!got_sync_mode) {
    mout << "Synchronous/Asynchronous unspecified, using default value of: "
         << "synchronous." << endl;
  }
  else {
    if(sync_mode)
      {
        fout << "Synchronous mode." << endl;
        if(loopback_edges)
          fout << "Will show loopback edges." << endl;
        else
          fout << "Will not show loopback edges." << endl;
      }
    else
      {
        fout << "Asynchronous mode." << endl;
        if(loopback_edges)
          {
            mout << "Note: -loopback_edges parameter will be ignored, "
                 << "it is only valid in synchronous mode" << endl;
                     
            loopback_edges = false;
          }
      }
  }
  if(!got_max_bits) {
    fout << "Maximum number of bits not specified, only " << max_bits
         << "-bit (or fewer) signals will be analyzed)." << endl;
  }
  else {
    fout << "Signals with up to " << max_bits << " bit(s) will be analyzed." << endl;
  }
  if(!got_modules) {
    mout << "A -modules line must be included specifying the names of "
         << "the modules to analyze." << endl;
    print_help();
    return false;
  }
  else {
    for(set<string>::iterator it = modules_to_analyze.begin();
        it != modules_to_analyze.end(); ++it) {
      fout << "Will analyze module " << *it << endl;
    }
    fout << "Main module: " << main_module << endl;
  }
  if(!got_signals) {
    fout << "No signals specified, will deduce from module list." << endl;
  }
  else {
    for(set<string>::iterator it = signals_to_analyze.begin();
        it != signals_to_analyze.end(); ++it) {
      fout << "Will analyze signal " << *it << endl;
    }
  }
  if(!got_transaction_mode) {
    mout << "Transaction/Non-transaction mode not specified, " 
         << "will try to find transactions." << endl;
  }
  else {
    if(transaction_mode)
      fout << "Will attempt to find transactions" << endl;
    else
      fout << "Will not attempt to find transactions" << endl;
  }             
  if(!got_start_transaction) {
    fout << "Transaction starts not specified, will attempt to "
         << "find transaction boundaries automatically." << endl;
  }
  else {
    for(unsigned int i = 0; i < start_signal_names.size(); ++i) {
      fout << "Will find transactions starting when " << start_signal_names[i] << " " << start_triggers[i] << "s\n";
    }
  }
  if(!got_verilog_file_names) {
    mout << "A -verilog line must be included to specify the verilog files "
         << "describing the modules to be analyzed -- so that we can "
         << "determine the inputs/outputs of the modules and the inter-module "
         << "connections." << endl;
    print_help();
    return false;
  }
  else {
    for(vector<string>::iterator it = verilog_f_names.begin();
        it != verilog_f_names.end(); ++it) {
      fout << "Will read verilog file " << *it << endl;
    }
  }
  if(!got_trace_file_name) {
    mout << "A -vcd line must be included to specify the trace file for "
         << "the simulation under analysis." << endl;
    print_help();
    return false;
  }
  else {
    fout << "Trace (vcd) file: " << trace_file_name << endl;
  }

  //If we only get one module to analyze and no signal list is specififed,
  //look at all its inputs/outputs (subject to max_bits restriction).
  if(modules_to_analyze.size() == 1 && !signal_mode) {
    signal_mode = true;
    all_signals = true;
  }

  return true;
}

void Config_Parser::print_help() {
  mout << "-------------------------------------------------------------------------------" << endl
       << "Possible settings (required unless otherwise specified -- I think): " << endl
       << "-------------------------------------------------------------------------------" << endl
       << "-name (name of directory to work in/file base name)" << endl
       << "-top (name of top module)" << endl
       << "-max_lines (number of lines to read from the trace file at once -- to control memory usage)" << endl
       << "-clock (name of the clock signal in the form module_name.signal_name)" << endl
       << "-reset (name of the reset signal in the form module_name.signal_name)" << endl
       << "-synchronous (true by default -- only sample signals at rising clock edges)" << endl
       << "-loopback_edges (if included, show when a state is held for more than one consecutive clock in synchronous mode)" << endl
       << "-asynchronous (sample signals whenever any signal changes)" << endl
       << "-max_bits (the maximum number of bits a signal can have and still be analyzed -- to exclude data)" << endl
       << "-modules (the module to be considered; assuming the signals to analyze are not directly specified, if one module is given here, all its signals will be analyzed, and if multiple modules are given only the signals at interfaces between two of the modules will be considered)" << endl
       << "-signals (an optional list of specific signals to analyze)" << endl
       << "-ignore_states (an optional list of states to ignore during analysis, in binary format)" << endl
       << "-show_hex (if included, state bits will appear as hex in the diagrams)" << endl
       << "-strip_leading_zeros (don't display leading zeros in transaction/state diagrams)" << endl
       << "-transaction_mode (true by default, if set to false, no attempt will be made to find transactions)" << endl
       << "-start_transaction (an optional specification of a condition which is considered to start a transaction, this option may be repeated multiple times in one file, in the form \"-start_transaction signal_name [01|10]\" where 01 signifies rising and 10 signifies falling" << endl
       << "-verilog (the list of files containing descriptions of the modules to analyze -- used to determine the inputs/outputs of each module and the inter-module connections)" << endl
       << "-vcd (the trace file to be used for analysis -- exactly one please)" << endl
       << endl;
}
