/**************************************************************************
 *   Inferno : An automatic semantic information extractor                 *
 *   Copyright (C) 2006-2010  University of Michigan                       *
 *   http://www.eecs.umich.edu/inferno                                     *
 *                                                                         *
 *   Contributors include Andrew DeOrio, Adam Bauserman,                   *
 *   Valeria Bertacco and Beth Isaksen.                                    *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA         *
 *   02110-1301  USA                                                       *
 *                                                                         *
 *   Andrew DeOrio <awdeorio@umich.edu>                                    *
 *   Valeria Bertacco <valeria@umich.edu>                                  *
 *                                                                         *
 *   University of Michigan                                                *
 *   Electrical Engineering and Computer Science Dept.                     *
 *   2260 Hayward St.                                                      *
 *   Ann Arbor, MI 48109-2121                                              *
 **************************************************************************/

#include "Dot_File_Printer.h"

const int Dot_File_Printer::num_colors = 9;
const int Dot_File_Printer::num_shapes = 5;
const int Dot_File_Printer::num_styles = 4;
const int Dot_File_Printer::num_edge_colors = 8;

// color schemes to use - make sure numbers above match the chosen schemes
const string bubble_color_scheme = "set19";
const string edge_color_scheme = "dark28";      
        
const string Dot_File_Printer::shapes[5] =
  {
    "ellipse",
    "box",
    "diamond",
    "hexagon",
    "doubleoctagon"
  };

const string Dot_File_Printer::styles[4] = 
  {
    "filled",
    "bold",
    "dotted",
    "dashed"
  };




Dot_File_Printer::Dot_File_Printer(){}

Dot_File_Printer::~Dot_File_Printer() {}

void Dot_File_Printer::print_state_graph(const string& file_name, const map<int,State>& states) const {

  if(states.size() == 0) {
    mout << "Error: cannot print state dot file because states has size 0" << endl;
    return;
  }
  ofstream out_file;
  out_file.open(file_name.c_str(),ofstream::out);
  if(!out_file.is_open()) {
    mout << "Unable to open file " << file_name << " for printing state dot file.\n";
    return;
  }
  else {
    fout << "Opened file " << file_name << " for printing state dot file.\n";
  }
  string key = nice_key(states.begin()->second);
  out_file << graph_header(file_name,key);

  vector<State> states_vector;
  for(map<int,State>::const_iterator it = states.begin(); it != states.end(); ++it) {
    states_vector.push_back(it->second);
  }
  out_file <<  bubbles(states_vector,0);

  set<pair<int,int> > seen_edges;
  map<int,State>::const_iterator to_state_it;

  // find total number of transitions
  int total_transitions=0;
  // iterate over bubbles
  for(map<int,State>::const_iterator from_state_it = states.begin(); from_state_it != states.end(); ++from_state_it) {
    // iterate over next bubble arrows
    for(vector< pair<int,long> >::const_iterator it = from_state_it->second.nexts.begin(); it != from_state_it->second.nexts.end(); ++it) {
      total_transitions = total_transitions + it->second;
    }
  }
        
  // find min and max
  float edge_weight=0;
  float edge_weight_max=0;
  float edge_weight_min=1;
  // iterate over bubbles
  for(map<int,State>::const_iterator from_state_it = states.begin(); from_state_it != states.end(); ++from_state_it) {
    // iterate over next bubble arrows
    for(vector< pair<int,long> >::const_iterator it = from_state_it->second.nexts.begin(); it != from_state_it->second.nexts.end(); ++it) {
      edge_weight = it->second / (float)total_transitions;
      if (edge_weight < edge_weight_min)
        edge_weight_min = edge_weight;
      if (edge_weight > edge_weight_max)
        edge_weight_max = edge_weight;
    }
  }

  // iterate over bubbles
  for(map<int,State>::const_iterator from_state_it = states.begin(); from_state_it != states.end(); ++from_state_it) {
    // iterate over next bubble arrows
    for(vector< pair<int,long> >::const_iterator it = from_state_it->second.nexts.begin(); it != from_state_it->second.nexts.end(); ++it) {
      to_state_it = states.find(it->first);

      // transitions on this arrow / total trans out of all bubbles
      edge_weight = it->second / (float)total_transitions;
      // transition on this arrow / total trans out of this bubble
      //it->second / (float)(from_state_it->second.total_weight)); // this is original from Adam
      out_file << edge(from_state_it->second,to_state_it->second,0,seen_edges,"/black","",edge_weight,edge_weight_min,edge_weight_max);
    }
  }
  out_file << "}" << endl;
  out_file.close();
}

void Dot_File_Printer::print_transaction_graph(const string& file_name, const vector<Transaction>& transactions,
                                               const map<int,State>& states) const {

  if(transactions.size() == 0) {
    mout << "Error: cannot print transaction dot file because transactions has size 0" << endl;
    return;
  }
  if(states.size() == 0) {
    mout << "Error: cannot print transaction dot file because states has size 0" << endl;
    return;
  }
        
  ofstream out_file;
  out_file.open(file_name.c_str(),ofstream::out);
  if(!out_file.is_open()) {
    mout << "Unable to open file " << file_name << " for printing transaction dot file.\n";
    return;
  }
  else {
    fout << "Opened file " << file_name << " for printing transaction dot file.\n";
  }

  string key = nice_key(states.begin()->second);
  out_file << graph_header(file_name,key);
        
  int transaction_ID = 0;
  for(vector<Transaction>::const_iterator it = transactions.begin(); it != transactions.end(); ++it) {
    out_file << transaction_sub_graph(*it,states,transaction_ID);
    ++transaction_ID;
  }
  out_file << "}" << endl;
  out_file.close();
}

//Return an appropriate key for a dot file
string Dot_File_Printer::nice_key(const State& state) const {
  map<string,FourValuedVector> values = state.values;
  map<string,string> module_results;
        
  string signal_name, module_name;
  for(state_values_t::const_iterator it = values.begin(); it != values.end(); ++it) {
    signal_name = it->first.substr(it->first.rfind(".") + 1);
    module_name = it->first.substr(0,it->first.rfind("."));

    if(module_results.find(module_name) == module_results.end()) {
      module_results[module_name] = signal_name;
    }
    else {
      module_results[module_name] += ", " + signal_name;
    }
  }
  //Break up the signals from different modules onto different lines,
  //and label the lines with the module names
  string result;
  for(map<string,string>::iterator it = module_results.begin(); it != module_results.end(); ++it) {
    result += table_row(table_cell(it->first + ":") + table_cell(it->second));
  }
  result = table(result);
  return result;
}

//Return an appropriate label for a dot file state bubble (for a "state" diagram)
string Dot_File_Printer::nice_label(const State& state) const {
  map<string,FourValuedVector> values = state.values;
  string module_name;
  map<string,string> module_results;
        
  for(state_values_t::const_iterator it = values.begin(); it != values.end(); ++it) {

    module_name = get_short_module_name(it->first.substr(0,it->first.rfind(".")));
    module_results[module_name] += table_cell(it->second.get_string_w_dont_cares(show_hex, strip_leading_zeros));

  }
  //Break up the signals from different modules onto different lines,
  //and label the lines with the module names
  string result;
  for(map<string,string>::iterator it = module_results.begin(); it != module_results.end(); ++it) {
    result += table_row(table_cell(it->first + ":") + it->second);
  }
  result = table(result);
  return result;
}

vector<State> Dot_File_Printer::find_transaction_states(const Transaction& t, const map<int,State>& states) const {
  vector<int> state_IDs = t.sequence;
  for(vector<Loop>::const_iterator loop_it = t.loops.begin(); loop_it != t.loops.end(); ++loop_it) {
    state_IDs.insert(state_IDs.end(),loop_it->sequence.begin(), loop_it->sequence.end());
  }
  vector<State> transaction_states;
  for(vector<int>::iterator it = state_IDs.begin(); it != state_IDs.end(); ++it) {
    map<int,State>::const_iterator sit = states.find(*it);
    if(sit == states.end()) {
      mout << "Error while printing transaction graph, state " << *it
           << " occurred in transaction but was not found in states." << endl;
      transaction_states.clear();
      return transaction_states;
    }
    if(find(transaction_states.begin(),transaction_states.end(),sit->second) == transaction_states.end()) {
      transaction_states.push_back(sit->second);
    }
  }
  return transaction_states;
}

string Dot_File_Printer::transaction_sub_graph_header(const Transaction& t, int sub_graph_ID) const {
  string result;
  result = "subgraph cluster" + itoa(sub_graph_ID) + " {\n";;
  result += "label=\"" + itoa(t.num_occurrences) + " times, first time: " + itoa(t.time_first_seen) + "\"\n";
  return result;
}

string Dot_File_Printer::bubbles(const vector<State>& states, int transaction_ID) const {

  string result, label;
  bool top_state;
  for(vector<State>::const_iterator it = states.begin(); it != states.end(); ++it) {
    if(it == states.begin()) {
      top_state = true;
    }
    else {
      top_state = false;
    }
    label = nice_label(*it);
    result += bubble(*it,transaction_ID,top_state,label);
  }
  return result;
}

string Dot_File_Printer::bubble(const State& state, int transaction_ID, bool top_state, const string& label) const {

  string result;
  if(top_state) {
    result +=  "{rank=\"source\"; \"" + itoa(state.ID) + "." + itoa(transaction_ID) + "\";}" + "\n";
  }
  string color = itoa((state.ID % num_colors) + 1);
  string shape = shapes[(state.ID/num_colors) % num_shapes];
  string style = styles[(state.ID/(num_shapes*num_colors)) % num_styles];
  result += itoa(state.ID) + "." + itoa(transaction_ID);
  result += " [color=\"" + color + "\" ";
  result += "colorscheme=\"" + bubble_color_scheme + "\" ";
  result += "style=\"" + style + "\" ";
  result += "shape=\"" + shape + "\" ";
  result += "label=\n" + label + "];\n";
  return result;
}

string Dot_File_Printer::edge(const State& from_state,
                              const State& to_state,
                              int transaction_ID,
                              set<pair<int,int> >& seen_edges,
                              const string& color,
                              const string& times_string,
                              const float& weight,
                              const float& min_weight,
                              const float& max_weight
                              ) const {
  string result;
  result += "\t" + itoa(from_state.ID) + "." + itoa(transaction_ID);
  result += " -> " + itoa(to_state.ID) + "." + itoa(transaction_ID) + " [label=\"" + times_string;
  if(seen_edges.find(make_pair(from_state.ID,to_state.ID)) == seen_edges.end()) {
    result += get_edge_label(from_state,to_state,weight);
    seen_edges.insert(make_pair(from_state.ID,to_state.ID));
  }
  result += "\" colorscheme=\"" + edge_color_scheme + "\" ";
  result += "fontcolor=\"" + color + "\" ";
  result += "color=\"" + color + "\" ";

  // set line widths, but only if fxn was called with max_weight set
  // this could mess up transaction diagrams otherwise
  if (max_weight != 0) {
    float linewidth=1; // range: 1 - 10
    float arrowsize=1; // range: 1 - 3
    // scale weight to linewidth and arrowsize ranges
    linewidth = (weight/max_weight)*9 + 1;
    arrowsize = (weight/max_weight)*2 + 1;

    // write strings to DOT file
    stringstream ss;
    ss << "style=\"setlinewidth(" << linewidth << ")\" ";
    result += ss.str();
    ss.flush();
    ss << "arrowsize=\"" << arrowsize << "\"";
    result += ss.str();
  }

  result += "]\n";
  return result;
}

string Dot_File_Printer::get_edge_label(const State& from_state, const State& to_state,
                                        const float& weight) const {

  vector<string> signal_names = from_state.get_signal_names();
  vector<FourValuedVector> from_values = from_state.get_signal_values(signal_names);
  vector<FourValuedVector> to_values = to_state.get_signal_values(signal_names);

  string edge_label;
  for(unsigned int i = 0; i < from_values.size(); ++i) {
    if(from_values[i] != FourValuedVector("0") && to_values[i] == FourValuedVector("0"))
      edge_label += "F " + get_really_short_name(signal_names[i]) + "\\n";
    else if(from_values[i] != FourValuedVector("1") && to_values[i] == FourValuedVector("1"))
      edge_label += "R " + get_really_short_name(signal_names[i]) + "\\n";
  }
  if(weight > 0) 
    {
      char weight_s[5];
                
      sprintf(weight_s, "%.3f", weight);
      edge_label += string(weight_s) + "\\n";
    }
  return edge_label;
}

string Dot_File_Printer::loop_times(const Loop& loop) const {
  string result;
  if(loop.min_times == loop.max_times) {
    if(loop.min_times == 1) {
      result = itoa(loop.min_times) + " time";
    }
    else {      
      result = itoa(loop.min_times) + " times";
    }
  }
  else {
    result = itoa(loop.min_times) + "-";
    result += itoa(loop.max_times) + " times";
  }
  return result + "\\n";
}

string Dot_File_Printer::transaction_sub_graph(const Transaction& t, const map<int,State>& states, int transaction_ID) const {
  vector<State> transaction_states = find_transaction_states(t, states);
        
  vector<int> cur_sequence = t.sequence;
  if(cur_sequence.size() < 1) {
    mout << "Error while printing dot file -- transaction had a sequence of size 0.\n";
    return "";
  }

  string result;
  result += transaction_sub_graph_header(t, transaction_ID);
  result += bubbles(transaction_states, transaction_ID);

  cur_sequence = t.sequence;
  set<pair<int,int> > seen_edges;
  string times_string;
  map<int,State>::const_iterator from_state_it, to_state_it;

  for(unsigned int i = 1; i < cur_sequence.size(); ++i) {
    from_state_it = states.find(cur_sequence[i-1]);
    to_state_it = states.find(cur_sequence[i]);
    result += edge(from_state_it->second,to_state_it->second,transaction_ID,seen_edges,
                   "/black",times_string,0);
  }
  int loop_number = 0;
  for(vector<Loop>::const_iterator loop_it = t.loops.begin(); loop_it != t.loops.end(); ++loop_it) {

    cur_sequence = loop_it->sequence;
    string color = itoa((loop_number % num_edge_colors) + 1);
    for(unsigned int i = 0; i < cur_sequence.size() - 1; ++i) {
      from_state_it = states.find(cur_sequence[i]);
      to_state_it = states.find(cur_sequence[i+1]);
      if(i == 0) {
        times_string = loop_times(*loop_it);
      }
      result += edge(from_state_it->second,to_state_it->second,transaction_ID,seen_edges,
                     color,times_string, 0);
    }
    ++loop_number;
  }
  result += "}\n\n";
  return result;
}

string Dot_File_Printer::graph_header(const string& file_name, const string& key) const {
  string result;
  result +="digraph \"" + file_name + "\" {\n";
  result +="size=\"7.5,10\";\n";
  //for landscape, comment out above size line and use the next two
  //result +="size=\"10,7.5\";\n";
  //result +="orientation=landscape\n\n";
  result +="label=" + key + "\n";
  return result;
}

string Dot_File_Printer::table(const string& contents) const {
  return "\t<<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\">\n" + contents + "\t</table>>";
}

string Dot_File_Printer::table_row(const string& contents) const {
  return "\t<tr>\n" + contents + "\t</tr>\n";
}

string Dot_File_Printer::table_cell(const string& contents) const {
  return "\t\t<td>" + contents + "</td>\n";
}

string Dot_File_Printer::font_color(const string& color, const string& contents) const {
  return "<font color=\"" + color + "\">" + contents + "</font>";
}
