/**************************************************************************
 *   Inferno : An automatic semantic information extractor                 *
 *   Copyright (C) 2006-2010  University of Michigan                       *
 *   http://www.eecs.umich.edu/inferno                                     *
 *                                                                         *
 *   Contributors include Andrew DeOrio, Adam Bauserman,                   *
 *   Valeria Bertacco and Beth Isaksen.                                    *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA         *
 *   02110-1301  USA                                                       *
 *                                                                         *
 *   Andrew DeOrio <awdeorio@umich.edu>                                    *
 *   Valeria Bertacco <valeria@umich.edu>                                  *
 *                                                                         *
 *   University of Michigan                                                *
 *   Electrical Engineering and Computer Science Dept.                     *
 *   2260 Hayward St.                                                      *
 *   Ann Arbor, MI 48109-2121                                              *
 **************************************************************************/

#include "Module.h"

Module::Module()
{
  type = "";
  inputs.clear();
  outputs.clear();
  internals.clear();
}

Module::~Module()
{
}

void Module::clear() {
  type = "";
  inputs.clear();
  outputs.clear();
  internals.clear();
  children.clear();
}

bool Module::has_child(const string& child_name) const {
  return children.find(child_name) != children.end();
}

signal_direction_t Module::get_direction(const string& sig_name) const {
  if(inputs.find(sig_name) != inputs.end())
    return INPUT;
  else if(outputs.find(sig_name) != outputs.end())
    return OUTPUT;
  else
    return INTERNAL;
}

vector<string> Module::get_interface(const string& module_name, vector<string>& signal_name,
                                     const string& main_module) const {
  children_t::const_iterator mit = children.find(module_name);
  if(mit == children.end()) {
    vector<string> results;
    results.push_back("");
    mout << "Error no child " << module_name << " found for " << type << endl;
    return results;
  }
  string signal_name_in = signal_name[0];
  signal_name.clear();
  string global_name = mit->second.get_global_name(signal_name_in);
  vector<string> results;
  for(children_t::const_iterator it = children.begin(); it != children.end(); ++it) {
    if(it->second.name != mit->second.name) {
      if(it->second.has_global_name(global_name)) {
        signal_name.push_back(it->second.get_local_name(global_name));
        results.push_back(it->second.name);
      }
    }
  }
  if(results.size() == 0)
    results.push_back("");
  return results;
}

string Module::get_child_type(const string& child_name) const {
  children_t::const_iterator it = children.find(child_name);
  return it->second.type;
}

vector<string> Module::get_output_signal_names(const vector<string>& signal_names) const {

  vector<string> output_signal_names;
        
  for(vector<string>::const_iterator it = signal_names.begin(); it != signal_names.end(); ++it) {
    string cur_signal = *it;
    //          mout << "signal " << get_really_short_name(cur_signal) << " has direction " << get_direction(get_really_short_name(cur_signal)) << endl;
    if(get_direction(get_really_short_name(cur_signal)) == OUTPUT) {
      output_signal_names.push_back(*it);
    }
  }
  return output_signal_names;
}

ostream& operator<<(ostream &os, const Module &m) {

  map<string,int>::const_iterator it;

  os << "Module: " << m.type << endl;
        
  os << "\tInputs:\n";
  for(it = m.inputs.begin(); it != m.inputs.end(); ++it) {
    os << "\t\t" << it->first << "\t[" << it->second << "]\n";
  }
  os << "\tOutputs:\n";
  for(it = m.outputs.begin(); it != m.outputs.end(); ++it) {
    os << "\t\t" << it->first << "\t[" << it->second << "]\n";
  }
  os << "\tChildren:\n";
  for(children_t::const_iterator sit = m.children.begin();
      sit != m.children.end();++sit) {
    os << sit->second << endl;
  }

  return os;
}


