/**************************************************************************
 *   Inferno : An automatic semantic information extractor                 *
 *   Copyright (C) 2006-2010  University of Michigan                       *
 *   http://www.eecs.umich.edu/inferno                                     *
 *                                                                         *
 *   Contributors include Andrew DeOrio, Adam Bauserman,                   *
 *   Valeria Bertacco and Beth Isaksen.                                    *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA         *
 *   02110-1301  USA                                                       *
 *                                                                         *
 *   Andrew DeOrio <awdeorio@umich.edu>                                    *
 *   Valeria Bertacco <valeria@umich.edu>                                  *
 *                                                                         *
 *   University of Michigan                                                *
 *   Electrical Engineering and Computer Science Dept.                     *
 *   2260 Hayward St.                                                      *
 *   Ann Arbor, MI 48109-2121                                              *
 **************************************************************************/

#include "Signal.h"

Signal::Signal() {}

Signal::Signal(const string& full_name_in, const string& name_in, const string& module_name_in,
               const string& parent_name_in, int num_bits_in):
  full_name(full_name_in), name(name_in), module_name(module_name_in), parent_name(parent_name_in),
  num_bits(num_bits_in), direction(INTERNAL) {}

Signal::~Signal() {}

vector<Signal> Signal::get_single_bits() const {
  vector<Signal> results;

  if(num_bits == 1) {
    results.push_back(*this);
    return results;
  }

  Signal cur_signal;
  for(int i = num_bits - 1; i >= 0; --i) {
    cur_signal = *this;
    cur_signal.num_bits = 1;
    cur_signal.full_name = full_name + "[" + itoa(i) + "]";
    values_t cur_values;
    for(values_t::const_iterator it = values.begin(); it != values.end(); ++it) {
      cur_values.push_back(make_pair(it->first,it->second[num_bits - i - 1]));
    }
    cur_signal.values = cur_values;
    results.push_back(cur_signal);
  }
  return results;
}

void Signal::remove_glitches() {
  values_t tmp_values = values;
  values.clear();

  unsigned int i = 1;
  while(i < tmp_values.size()) {
    if(tmp_values[i].first != tmp_values[i-1].first
       && (i == 1 || tmp_values[i-1].second != values[values.size() - 1].second))
      values.push_back(tmp_values[i-1]);
    ++i;
  }
  if(tmp_values.size() > 0)
    values.push_back(tmp_values[tmp_values.size() - 1]);
}

void Signal::setup_iterators() {
  cur_it = values.begin();

  values_t::const_iterator tmp_it = cur_it;
  max_it = cur_it;

  if(tmp_it != values.end())
    ++tmp_it;
  while(tmp_it != values.end()) {
    ++tmp_it;
    ++max_it;
  }
        
}

//return true if done
bool Signal::goto_next(unsigned long min_time) {
  if(cur_it == max_it || min_time >= max_it->first) {
    return true;
  }
  values_t::const_iterator tmp_it = cur_it;
  ++tmp_it;
  while(tmp_it != max_it && tmp_it->first <= min_time) {
    ++tmp_it;
    ++cur_it;
  }
  if(cur_it != max_it && tmp_it->first <= min_time)
    ++cur_it;
  return cur_it == max_it;
}

FourValuedVector Signal::cur_value() const {
  if(cur_it == values.end()) {
    mout << "Error: empty signal: " << full_name << "\n";
    return FourValuedVector("X");
  }
  return cur_it->second;
}

unsigned long Signal::next_pos_edge(unsigned long last_time) {
  if(cur_it == max_it) {
    if(cur_it == values.end()) {
      mout << "Error: empty signal.\n";
      return 0;
    }
    return cur_it->first;
  }
  //    mout << "next_posedge, " << full_name << " last_time: " << last_time << " cur_value: " << cur_it->second << " cur_time: " << cur_it->first << endl;
  while(cur_it != values.end() && cur_it->first <= last_time) {
    ++cur_it;
  }
  while(cur_it != values.end() && cur_it->second != FourValuedVector("1"))
    ++cur_it;
  if(cur_it == values.end())
    --cur_it;
  return cur_it->first;
}

unsigned long Signal::prev_pos_edge(unsigned long last_time) {
  if(cur_it == values.begin()) {
    if(cur_it == values.end()) {
      mout << "Error: empty signal.\n";
      return 0;
    }
    return cur_it->first;
  }
  goto_next(last_time);
  values_t::const_iterator tmp_it = cur_it;
  while(tmp_it != values.begin() && tmp_it->first >= last_time) {
    --tmp_it;
  }
  while(tmp_it != values.begin() && tmp_it->second != FourValuedVector("1"))
    --tmp_it;
  if(tmp_it != values.begin())
    --tmp_it;
  return tmp_it->first;
}

unsigned long Signal::next_time(unsigned long last_time) const {
  if(cur_it == max_it) {
    if(cur_it == values.end()) { 
      mout << "Error: empty signal.\n";
      return 0;
    }
    return cur_it->first;
  }
  values_t::const_iterator tmp_it = cur_it;
  while(tmp_it != values.end() && tmp_it->first <= last_time) {
    ++tmp_it;
  }
  return tmp_it->first;
}

unsigned long Signal::max_time() const {
  if(values.size() == 0)
    return 0;
  else
    return values.back().first;
}

void Signal::replace_values(const values_t& new_values) {
  if(values.size() > 1) {
    values_t::iterator it = values.end();
    --it;
    values.erase(values.begin(),it);
  }
  values.insert(values.end(),new_values.begin(),new_values.end());
}

void Signal::clear_values() {
  values.clear();
}

bool Signal::not_internal() {
  return (direction == INPUT || direction == OUTPUT);
}

ostream& operator<<(ostream &os, const Signal &s) {
  os << s.full_name << " " << s.num_bits;
  if(s.direction == INPUT)
    mout << " input\n";
  else if(s.direction == OUTPUT)
    mout << " output\n";
  else
    mout << " internal\n";
  for(values_t::const_iterator it = s.values.begin(); it != s.values.end(); ++it) {
    os << it->first << ": " << it->second << endl;
  }
  return os;
}

