/**************************************************************************
 *   Inferno : An automatic semantic information extractor                 *
 *   Copyright (C) 2006-2010  University of Michigan                       *
 *   http://www.eecs.umich.edu/inferno                                     *
 *                                                                         *
 *   Contributors include Andrew DeOrio, Adam Bauserman,                   *
 *   Valeria Bertacco and Beth Isaksen.                                    *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA         *
 *   02110-1301  USA                                                       *
 *                                                                         *
 *   Andrew DeOrio <awdeorio@umich.edu>                                    *
 *   Valeria Bertacco <valeria@umich.edu>                                  *
 *                                                                         *
 *   University of Michigan                                                *
 *   Electrical Engineering and Computer Science Dept.                     *
 *   2260 Hayward St.                                                      *
 *   Ann Arbor, MI 48109-2121                                              *
 **************************************************************************/

#include "Signal_Group.h"

Signal_Group::Signal_Group() : have_clock(false), have_reset(false) {}
Signal_Group::~Signal_Group() {}

void Signal_Group::configure(const string& clock_name_in, const string& reset_name_in, bool sync_mode_in) {
  clock_name = clock_name_in;
  reset_name = reset_name_in;
  sync_mode = sync_mode_in;
}

void Signal_Group::add_signal(const Signal& s) {
  if(s.full_name == clock_name) {
    clock_signal = s;
    have_clock = true;
  }
  else if(s.full_name == reset_name) {
    reset_signal = s;
    have_reset = true;
  }
  else {
    signals[s.full_name] = s;
  }
}

void Signal_Group::prepare_to_find_states(unsigned long max_time) {
  add_ends(max_time);
  restart();
  if(have_reset)
    reset_signal.setup_iterators();
  if(have_clock)
    clock_signal.setup_iterators();
}

void Signal_Group::replace_values(const string& signal_name, const values_t& values) {
  //    cout << "replace_values on " << signal_name << ", size = " << values.size() << endl ;
  if(signal_name == clock_name)
    clock_signal.replace_values(values);
  else if(signal_name == reset_name)
    reset_signal.replace_values(values);
  else
    signals[signal_name].replace_values(values);
}

int Signal_Group::num_signals() const {
  return signals.size();
}

void Signal_Group::restart() {
  for(signals_t::iterator it = signals.begin(); it != signals.end(); ++it)
    it->second.setup_iterators();
  num_done = 0;
}

void Signal_Group::goto_next(unsigned long min_next_time) {
  num_done = 0;
  for(signals_t::iterator it = signals.begin(); it != signals.end(); ++it) {
    if(it->second.goto_next(min_next_time))
      ++num_done;
  }
}

unsigned long Signal_Group::first_min_next_time() {
  unsigned long min_next_time = 0;
  values_t::iterator it = reset_signal.values.begin();
  while(it != reset_signal.values.end() && it->second != FourValuedVector("0")) {
    min_next_time = it->first;
    ++it;
  }
  if(sync_mode) {
    clock_signal.goto_next(min_next_time);
    min_next_time = clock_signal.prev_pos_edge(min_next_time);
    min_next_time = clock_signal.prev_pos_edge(min_next_time - 1);
  }
  return min_next_time;
}

unsigned long Signal_Group::min_next_time(unsigned long last_time) {
  unsigned long cur_next_time=0, min_next_time=0;
  bool initializing = true;
        
  for(signals_t::iterator it = signals.begin(); it != signals.end(); ++it){
    if(it->second.cur_it->first < last_time)
      it->second.goto_next(last_time);
  }
  for(signals_t::const_iterator it = signals.begin(); it != signals.end(); ++it){
    cur_next_time = it->second.next_time(last_time);
    if(initializing || cur_next_time < min_next_time) {
      min_next_time = cur_next_time;
      initializing = false;
    }
  }
  if(sync_mode) {
    unsigned long pos_edge;
    pos_edge = clock_signal.next_pos_edge(min_next_time);
    if(pos_edge > min_next_time)
      min_next_time = pos_edge;
    else
      min_next_time = max_time();
    //          cout << "min_next_time after clock: " << min_next_time << endl;
  }
  return min_next_time;
}

unsigned long Signal_Group::max_time() const {
  unsigned long cur_time=0, max_time=0;
  bool initializing = true;
  for(signals_t::const_iterator it = signals.begin(); it != signals.end(); ++it){
    cur_time = it->second.max_time();
    if(initializing || cur_time > max_time) {
      max_time = cur_time;
      initializing = false;
    }
  }
  return max_time;
}

void Signal_Group::add_ends(unsigned long max_t) {
  for(signals_t::iterator it = signals.begin(); it != signals.end(); ++it){
    if(it->second.max_time() != max_t || it->second.values.size() == 0) {
      if(it->second.values.size() == 0) {
        it->second.values.push_back(make_pair(0,FourValuedVector("X")));
        if(max_t != 0)
          it->second.values.push_back(make_pair(max_t,FourValuedVector("X")));
      }
      else {
        int num_values = it->second.values.size();
        if(num_values > 1 && it->second.values[num_values - 2].second == it->second.values[num_values - 1].second)
          it->second.values[num_values - 1].first = max_t;
        else
          it->second.values.push_back(make_pair(max_t,it->second.values.back().second));
      }
    }
  }
}

map<string,FourValuedVector> Signal_Group::get_values() const {

  map<string,FourValuedVector> values;
  for(signals_t::const_iterator it = signals.begin(); it != signals.end(); ++it)
    values[it->first] = it->second.cur_value();
  return values;
}

bool Signal_Group::all_done() const {
  return (num_done == signals.size());
}

