/**************************************************************************
 *   Inferno : An automatic semantic information extractor                 *
 *   Copyright (C) 2006-2010  University of Michigan                       *
 *   http://www.eecs.umich.edu/inferno                                     *
 *                                                                         *
 *   Contributors include Andrew DeOrio, Adam Bauserman,                   *
 *   Valeria Bertacco and Beth Isaksen.                                    *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA         *
 *   02110-1301  USA                                                       *
 *                                                                         *
 *   Andrew DeOrio <awdeorio@umich.edu>                                    *
 *   Valeria Bertacco <valeria@umich.edu>                                  *
 *                                                                         *
 *   University of Michigan                                                *
 *   Electrical Engineering and Computer Science Dept.                     *
 *   2260 Hayward St.                                                      *
 *   Ann Arbor, MI 48109-2121                                              *
 **************************************************************************/

#include "State.h"

State::State()
{
  total_weight = 0;
}

State::State(int ID_in,const state_values_t& values_in): ID(ID_in), values(values_in)
{
  total_weight = 0;
}

State::~State()
{
}

//Add a next state ID to the list (if it's not already there)
void State::add_next(int next_ID) {
        
  vector<next_state_t>::iterator it; // = find(nexts.begin(),nexts.end(),next_ID);
        
  for(it = nexts.begin(); it != nexts.end(); ++it)
    if(it->first == next_ID) break;
        
  if(it == nexts.end())
    nexts.push_back(next_state_t(next_ID, 1));
  else
    it->second++;
                
  total_weight++;
}


//Remove a signal -- called if we determine the signal is always constant
void State::remove_signal(const string& signal_name) {
  if(values.find(signal_name) != values.end())
    values.erase(signal_name);
}

//If all signals were removed, then they must have all been constant
bool State::all_constant() const {
  return (values.size() == 0);
}

//Return the total number of bits in the signals specified
int State::get_num_signal_bits(const vector<string>& signal_names) const {
  int num_bits = 0;
  for(vector<string>::const_iterator it = signal_names.begin(); it != signal_names.end(); ++it) {
    state_values_t::const_iterator vit = values.find(*it);
    //If we don't have the signal, print an error message and return 0
    if(values.find(*it) == values.end()) {
      mout << "Error: requested number of bits in non-existent signal " << *it << " from state " << ID << endl;
      return 0;
    }
    num_bits += vit->second.size();
  }
  return num_bits;
}

//Return the number of bits in the given signal
int State::get_num_signal_bits(const string& signal_name) const {
  state_values_t::const_iterator vit = values.find(signal_name);
  //If we don't have the signal, print an error message and return 0
  if(vit == values.end()) {
    mout << "Error: requested number of bits in non-existent signal " << signal_name << " from state " << ID << endl;
    return 0;
  }
  return vit->second.size();
}

//Return the value of the given signal in this state
FourValuedVector State::get_signal_value(const string& signal_name) const {
  state_values_t::const_iterator it = values.find(signal_name);
  //If we don't have the signal, print an error message and return 0
  if(it == values.end()) {
    mout << "Error: requested value of non-existent signal " << signal_name << " from state " << ID << endl;
    return FourValuedVector("0");
  }
  return it->second;
}

//Return a map from the names of the given signals to their values in this state
map<string,FourValuedVector> State::get_name_value_map(const vector<string>& signal_names) const {
  map<string,FourValuedVector> results;
  for(vector<string>::const_iterator it = signal_names.begin(); it != signal_names.end(); ++it) {
    //If we don't have the signal, print an error message and return empty map
    if(values.find(*it) == values.end()) {
      mout << "Error: requested name-value map of non-existent signal " << *it << " from state " << ID << endl;
      results.clear();
      return results;
    }
    results[*it] = get_signal_value(*it);
  }
  return results;
}

//Return a string with one character showing the value of each of the signals specified
string State::get_signal_values_string(const vector<string>& signal_names) const {
  string result;
  string cur_values;
  for(vector<string>::const_iterator it = signal_names.begin(); it != signal_names.end(); ++it) {
    //If we don't have the signal, print an error message and return empty string
    if(values.find(*it) == values.end()) {
      mout << "Error: requested signal/values string of non-existent signal " << *it << " from state " << ID << endl;
      return "";
    }
    // get value in binary, without stripping leading zeros
    cur_values = get_signal_value(*it).get_string_w_dont_cares(false, false);
    result += cur_values;
  }
  return result;
}

//Return a vector of values for the signals specified
vector<FourValuedVector> State::get_signal_values(const vector<string>& signal_names) const {
  vector<FourValuedVector> results;
  for(vector<string>::const_iterator it = signal_names.begin(); it != signal_names.end(); ++it) {
    //If we don't have the signal, print an error message and return empty vector
    if(values.find(*it) == values.end()) {
      mout << "Error: requested signal/values string of non-existent signal " << *it << " from state " << ID << endl;
      results.clear();
      return results;
    }
    results.push_back(get_signal_value(*it));
  }
  return results;
}

//Return a vector of the names of the signals
vector<string> State::get_signal_names() const {
  vector<string> signal_names;
  for(map<string,FourValuedVector>::const_iterator it = values.begin(); it != values.end(); ++it)
    signal_names.push_back(it->first);
  return signal_names;
}

//== is true if all the values are equal
bool State::operator==(const State& s) const {
  return (values == s.values);
}

ostream& operator<<(ostream &os, const State &s) {

  os << "State: " << s.ID << endl;

  for(state_values_t::const_iterator it = s.values.begin();
      it != s.values.end(); ++it) {
    os << it->first << "\t" << it->second << endl;
  }

  //    os << endl << "Nexts:" << endl;
  //    for(nexts_t::const_iterator it = s.nexts.begin();
  //            it != s.nexts.end(); ++it) {
  //            os << *it << endl;
  //    }
  return os;
}
