/**************************************************************************
 *   Inferno : An automatic semantic information extractor                 *
 *   Copyright (C) 2006-2010  University of Michigan                       *
 *   http://www.eecs.umich.edu/inferno                                     *
 *                                                                         *
 *   Contributors include Andrew DeOrio, Adam Bauserman,                   *
 *   Valeria Bertacco and Beth Isaksen.                                    *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA         *
 *   02110-1301  USA                                                       *
 *                                                                         *
 *   Andrew DeOrio <awdeorio@umich.edu>                                    *
 *   Valeria Bertacco <valeria@umich.edu>                                  *
 *                                                                         *
 *   University of Michigan                                                *
 *   Electrical Engineering and Computer Science Dept.                     *
 *   2260 Hayward St.                                                      *
 *   Ann Arbor, MI 48109-2121                                              *
 **************************************************************************/

#include "System.h"

System::System(){}

System::System(const string& config_file_name_in) : config_file_name(config_file_name_in),
                                                    done_parsing(false), last_ID(0), 
                                                    last_time(1) {}

System::System(const System& sys)
{   
  found_sis = sys.found_sis;
  found_abc = sys.found_abc;
    
  output_file_base = sys.output_file_base;
  top_name = sys.top_name;
  max_lines = sys.max_lines;
        
  reset_name = sys.reset_name;
  clock_name = sys.clock_name;
    
  transaction_mode = sys.transaction_mode;
  sync_mode = sys.sync_mode;
  signal_mode = sys.signal_mode;
  all_signals = sys.all_signals;
  start_mode = sys.start_mode;
        
  start_signal_names = sys.start_signal_names;
  start_triggers = sys.start_triggers;
  verilog_f_names = sys.verilog_f_names;
        
        
  trace_file_name = sys.trace_file_name;
  max_bits = sys.max_bits;  //max bits for a signal to be considered
  main_module = sys.main_module;
  set<string> signals_to_analyze, modules_to_analyze;
        
  //From reading verilog
  modules = sys.modules;
  module_name_to_type_translation = sys.module_name_to_type_translation;
  defined_constants = sys.defined_constants;

  //From reading trace file
  // don't copy this:
  //VCD_Parser vcd_parser;
        
  done_parsing = sys.done_parsing;

  states = sys.states;
  state_sequence = sys.state_sequence;
  time_sequence = sys.time_sequence;
  transactions = sys.transactions; //number of occurrences, sequence
  constants = sys.constants;
  last_ID = sys.last_ID;
  last_time = sys.last_time;

  //Signal_Group
  signals = sys.signals;

  //Dot_File_Printer df_printer;
}

System::~System() {}

bool System::read_config_file() {
  Config_Parser config_parser(config_file_name);
  bool success = config_parser.parse(output_file_base, top_name, max_lines, reset_name,
                                     clock_name, transaction_mode, sync_mode, signal_mode,
                                     all_signals, start_mode, start_signal_names,
                                     start_triggers, verilog_f_names, 
                                     trace_file_name, max_bits, main_module,
                                     signals_to_analyze, modules_to_analyze,
                                     loopback_edges, ignore_states, show_hex,
                                     strip_leading_zeros);
  signals.configure(clock_name,reset_name,sync_mode);
        
  df_printer.show_hex = show_hex;
  df_printer.strip_leading_zeros = strip_leading_zeros;
  return success;
}

bool System::parse_verilog() {
  //exceptional case -- the top level module has no name
  module_name_to_type_translation[top_name] = top_name; 

  Verilog_Parser vlog_parser(verilog_f_names,top_name);
  if(!vlog_parser.parse(modules, module_name_to_type_translation, defined_constants)) {
    mout << "Unable to parse verilog\n";
    return false;
  }
  for(set<string>::iterator it = modules_to_analyze.begin(); it != modules_to_analyze.end(); ++it) {
    string module_name = (*it);
                
    // get the shortened module name, i.e. 'parent.module'
    module_name = get_short_module_name(module_name);

    // translate parent module instance name to module type
    for(map<string,string>::const_iterator mit = module_name_to_type_translation.begin();
        mit != module_name_to_type_translation.end(); ++mit) {
      if(substr_before_dot(module_name) == substr_after_dot(mit->first)) {
        module_name = substr_before_dot(mit->second) + string(".") +
          substr_after_dot(module_name);
        cout << "new module name: " << module_name << endl;
        break;
      }             
    }


    if(module_name_to_type_translation.find(module_name) == module_name_to_type_translation.end()) {
      mout << "Error: no module instance found in verilog for " << module_name << endl;
      mout << "Found modules: " << endl;
      for(map<string,string>::const_iterator mit = module_name_to_type_translation.begin();
          mit != module_name_to_type_translation.end(); ++mit)
        mout << "name: " << mit->first << " type: " << mit->second << endl;
      return false;
    }
  }
  return true;
}

bool System::parse_beginning_of_trace_file() {
  //    fout << "System::parse_beginning_of_trace_file" << endl;
  vcd_parser.initialize(max_lines, max_bits, trace_file_name, modules_to_analyze,
                        clock_name, signal_mode, all_signals, signals_to_analyze,
                        sync_mode, ignore_states);
  if(!vcd_parser.parseBegin()) {
    mout << "Unable to parse beginning of file " << trace_file_name << "\n";
    return false;
  }
  //Determine which signals we actually want and assign them to the correct modules
  data_t tmp_data = vcd_parser.data, data;
  map<string,string> var_translation;

  fout << "Initial var translation contains " << vcd_parser.initial_var_translation.size() << " signals." << endl;
  for(unsigned int i = 0; i < vcd_parser.initial_var_translation.size(); ++i) {
    string symbol = vcd_parser.initial_var_translation[i].first;
    string full_name = vcd_parser.initial_var_translation[i].second;
    fout << "full_name: " << full_name << "(signal_mode: " << signal_mode << ")" << endl;
    int num_bits = vcd_parser.initial_num_bits[i];
    string name, module_name, parent_name, full_module_name;
    if(!parse_signal_name(full_name, name, module_name, parent_name, full_module_name)) {
      mout << "Failed to parse signal name: " << full_name << endl;
      return false;
    }
    //          mout << "parsed" << endl;
    Signal s = Signal(full_name,name,module_name,parent_name,num_bits);
    if((signal_mode && (all_signals || signals_to_analyze.find(full_name) != signals_to_analyze.end())) || (!signal_mode && re_translate(s))) {
                        
      fout << "changed name to: " << s.full_name << " with symbol: " << symbol << endl;
      //if(modules_to_analyze.find(parent_name + "." + module_name) == modules_to_analyze.end()) {
      cout << "module name: " << full_module_name << endl;
      if(modules_to_analyze.find(full_module_name) == modules_to_analyze.end()) {
        //if(0) {
        fout << "Warning: signal " << s.full_name << " is not from a module to be analyzed." << endl;
        //return false;
      }
      else
        {
          var_translation[symbol] = s.full_name;
          data[s.full_name] = s;
          signals.add_signal(s);
        }
    }
    else {
      //                        mout << "re_translate returned false" << endl;
    }
  }
  vcd_parser.data = data;
  vcd_parser.var_translation = var_translation;
        
  fout << "*** Final translation contains " << var_translation.size() << " signals:\n";

  for(map<string,string>::iterator it = var_translation.begin(); it != var_translation.end(); ++it)
    fout << "final translation from " << it->first << " to " << it ->second << endl;
  if(sync_mode && !signals.have_clock) {
    mout << "Error: synchronous mode requested but clock signal not found." << endl;
    return false;
  }
  if(!signals.have_reset) {
    mout << "Warning: reset signal not found (reset_name: " << reset_name << ")." << endl;
  }
  return true;
}

bool System::parse_more_of_trace_file() {
  if(!vcd_parser.parseContinue(modules, module_name_to_type_translation)) {
    mout << "Unable to parse more of file " << trace_file_name << "\n";
    return false;
  }
  fout << "Parsed more of file " << trace_file_name << "\n";
  done_parsing = vcd_parser.done_parsing;

  for(data_t::iterator it = vcd_parser.data.begin(); it != vcd_parser.data.end(); ++it) {
    signals.replace_values(it->first,it->second.values);
    it->second.clear_values();
  }
  fout << "Cleared data.\n";

  find_states();

  return true;
}

string System::get_interface(const string& parent_name, const string& child_name,
                             string& signal_name, const string& main_module) const {
  modules_t::const_iterator parent_it = modules.find(parent_name);
  //    mout << "parent module: " << parent_it->first << endl;
  vector<string> signal_names;
  signal_names.push_back(signal_name);
  //    mout << "child_name: " << child_name << " signal_name: " << signal_name << " main_module: " << main_module << endl;
  vector<string> neighbors = parent_it->second.get_interface(child_name,signal_names,main_module);
  string neighbor_type_name;
  signal_direction_t neighbor_direction;
  if(neighbors[0] == "") {
    signal_name = "";
    return neighbors[0];
  }
  for(unsigned int i = 0; i < neighbors.size(); ++i) {
    //          mout << "neighbors[" << i << "]: " << neighbors[i] << " signal_names[" << i << "]: " << signal_names[i] << endl;
    neighbor_type_name = parent_it->second.get_child_type(neighbors[i]);
    modules_t::const_iterator mit = modules.find(neighbor_type_name);
    neighbor_direction = mit->second.get_direction(signal_names[i]);
    if(neighbor_direction == OUTPUT) {
      signal_name = signal_names[i];
      return neighbors[i];
    }
  }
  signal_name = "";
  return neighbors[0];
}

//figure out which module we really want to treat the signal as part of, etc.
bool System::re_translate(Signal& s) {

  mout << "system::re_translate: " << s.full_name << " parent_name: " << s.parent_name << endl;
        
  map<string,string>::const_iterator parent_type_it = module_name_to_type_translation.find(s.parent_name);
  if(parent_type_it == module_name_to_type_translation.end()) {
    mout << "Error: failed to find module type for parent." << endl;
    mout << "module_name_to_type_translation: " << endl;
    for(map<string,string>::const_iterator it = module_name_to_type_translation.begin(); it != module_name_to_type_translation.end(); ++it) {
      mout << "name: " << it->first << " type: " << it->second << endl;
    }
  }
  assert(parent_type_it != module_name_to_type_translation.end());
  modules_t::const_iterator parent_it = modules.find(parent_type_it->second);
  assert(parent_it != modules.end());

  if(parent_it == modules.end() || !(parent_it->second.has_child(s.module_name))) {
    mout << "No module definition found with type " << s.parent_name << " and child " << s.module_name << endl;
    return false;
  }
  string type_name = parent_it->second.get_child_type(s.module_name);
  fout << "type_name: " << type_name << endl;
  modules_t::const_iterator it = modules.find(type_name);
  if(it == modules.end()) {
    mout << "No module definition found for " << type_name << endl;
    return false;
  }

  string neighbor_signal_name = s.name;
  string neighbor_name = get_interface(parent_type_it->second,s.module_name,
                                       neighbor_signal_name,main_module);
  string neighbor_type_name;
  signal_direction_t neighbor_direction = INTERNAL;
  if(neighbor_name != "") {
    neighbor_type_name = parent_it->second.get_child_type(neighbor_name);
    neighbor_direction = modules[neighbor_type_name].get_direction(neighbor_signal_name);
  }
  fout << "neighbor_name: " << neighbor_name << endl;
  fout << "neighbor_signal_name: " << neighbor_signal_name << endl;
  fout << "neighbor_direction: " << neighbor_direction << endl;
  if(signal_mode || (reset_name == s.full_name) || (clock_name == s.full_name)
     || (reset_name == s.parent_name + "." + neighbor_name + "." + neighbor_signal_name) 
     || (clock_name == s.parent_name + "." + neighbor_name + "." + neighbor_signal_name) ) {
    s.direction = OUTPUT;
    fout << "signal_mode/clock/reset" << endl;
    return true;
  }
  if(modules_to_analyze.find(s.parent_name + "." + s.module_name) == modules_to_analyze.end()
     || modules_to_analyze.find(s.parent_name + "." + neighbor_name) == modules_to_analyze.end()) {
    return false;
  }
  if(neighbor_direction == OUTPUT) {
    string name_front = s.full_name.substr(0,s.full_name.rfind("."));
    name_front = name_front.substr(0,name_front.rfind("."));                    
    s.full_name = name_front + "." + neighbor_name + "." + neighbor_signal_name;
    s.module_name = neighbor_name;
    s.name = neighbor_signal_name;
    s.direction = neighbor_direction;
    fout << "fixed signal" << endl;
    return true;
  }
  else if(s.direction == OUTPUT) {
    return true;
  }
  fout << "re_translated" << endl;
  return false;
}

void System::add_state(map<string,FourValuedVector> values, unsigned long cur_time) {
  int ID = states.size();
  unsigned int cur_ID;

  State s = State(ID,values);
  states_t::iterator sit = states.begin();
  while(sit != states.end()) {
    if(sit->second == s)
      break;
    ++sit;
  }
  if(sit != states.end()) {
    cur_ID = sit->second.ID;
  }
  else {
    states[ID] = s;
    cur_ID = ID;
    ++ID;
  }

  if(states.size() > 1 && (loopback_edges || (last_ID != cur_ID))) {
    states[last_ID].add_next(cur_ID);
  }
  if(loopback_edges || (last_ID != cur_ID)) {
    time_sequence.push_back(cur_time);
    state_sequence.push_back(cur_ID);
  }
  last_ID = cur_ID;
}

void System::find_states() {
  unsigned long min_next_time, max_time = signals.max_time();
        
  signals.prepare_to_find_states(max_time);
  min_next_time = signals.first_min_next_time();
  while(min_next_time < max_time) {
    signals.goto_next(min_next_time);
    add_state(signals.get_values(), min_next_time);
    last_time = min_next_time;
    min_next_time = signals.min_next_time(last_time);
  }
}

void System::find_constants_in_set(const vector<int>& state_IDs, map<string,FourValuedVector>& constants) {
  vector<string> constant_names = states[0].get_signal_names(), tmp_constant_names;
  map<string,FourValuedVector> orig_map = states[0].get_name_value_map(constant_names), cur_map;

  //for all states whose ID is in state_IDs, see if the value of each signal is always the same
  for(states_t::iterator it = states.begin(); it != states.end(); ++it) {
    if(find(state_IDs.begin(), state_IDs.end(),it->first) != state_IDs.end()) {
      cur_map = it->second.get_name_value_map(constant_names);
      tmp_constant_names = constant_names;
      constant_names.clear();
      for(unsigned int i = 0; i < tmp_constant_names.size(); ++i) {
        if(cur_map[tmp_constant_names[i]] == orig_map[tmp_constant_names[i]]) {
          constant_names.push_back(tmp_constant_names[i]);
        }
      }
    }
  }
  constants = states[0].get_name_value_map(constant_names);
}

void System::find_constants() {

  constants.clear();
  //figure out what is constant
  vector<int> state_IDs;
  for(states_t::iterator it = states.begin();it != states.end(); ++it) {
    state_IDs.push_back(it->first);
  }
  find_constants_in_set(state_IDs, constants);
  //move them out of the states
  for(states_t::iterator sit = states.begin(); sit != states.end(); ++sit) {
    for(map<string,FourValuedVector>::iterator cit = constants.begin(); cit != constants.end(); ++cit) {
      sit->second.remove_signal(cit->first);
    }
  }
  //remove states with no non-constants
  states_t tmp_states = states;
  states.clear();
  for(states_t::iterator sit = tmp_states.begin(); sit != tmp_states.end(); ++sit) {
    if(!(sit->second.all_constant())) {
      states.insert(*sit);
    }
  }
}

void System::find_transaction_constants() {
  map<string,FourValuedVector> constants;
  vector<int> state_IDs;
        
  for(vector<Transaction>::iterator it = transactions.begin();
      it != transactions.end(); ++it) {
    state_IDs = it->get_all_IDs();
    find_constants_in_set(state_IDs,constants);
    for(map<string,FourValuedVector>::iterator cit = constants.begin();
        cit != constants.end(); ++cit) {
      it->add_constant(cit->first);
    }
  }
}

void System::print_state_diagram() const {

  //    fout << "System::print_state_diagram" << endl;
  string file_name = output_file_base + "state.dot";
  df_printer.print_state_graph(file_name,states);
}

// returns false if no transactions printed to file
// (called with all_transactions=true when printing final transaction diagram)
bool System::print_dot_file(const string& dot_file_name, bool all_transactions) {

  //    fout << "System::print_dot_file" << endl;
  string actual_file_name;
  if(dot_file_name == "")
    actual_file_name = output_file_base + "transactions.dot";
  else
    actual_file_name = dot_file_name;
        
  if(INCREMENTAL_MODE && !all_transactions)
    {
      // remove transactions that were presented in
      // the previous iteration
      vector<Transaction>::iterator find_it;
      vector<Transaction> disp_transactions;
      int new_count = 0;
                
      for(vector<Transaction>::iterator it = transactions.begin();
          it != transactions.end(); ++it) 
        {
          // search for the transaction in approved transactions
          find_it = find(approved_transactions.begin(), approved_transactions.end(), (*it));
                                
          // didn't find it
          if(find_it == approved_transactions.end()) 
            {
              new_count++;
                                
                                  
              disp_transactions.push_back((*it));
              // add to approved transactions (may want to do this later)
              approved_transactions.push_back((*it));
            }
        }
                
      mout << new_count << " new transactions." << endl;
                
      if(disp_transactions.empty())
        {
          return false;
        }
      else
        {
          df_printer.print_transaction_graph(actual_file_name,disp_transactions,states);
          return true;
        }
    }
  else
    {
      // display all transactions
      df_printer.print_transaction_graph(actual_file_name,transactions,states);
      return !transactions.empty();
    }
}

bool System::start_transition(int last_state,int cur_state,const vector<string>& start_signals,
                              const vector<string>& start_triggers)
{
  //    fout << "System::start_transition\n";
  for(unsigned int i = 0; i < start_signals.size(); ++i) {
    string start_signal = start_signals[i];
    string start_trigger = start_triggers[i];
    fout << "checking " << start_signal << endl;
    FourValuedVector last_val = states[last_state].get_signal_value(start_signal);
    FourValuedVector cur_val = states[cur_state].get_signal_value(start_signal);
    fout << "last value: " << last_val << " current value: " << cur_val << endl;
    if(start_trigger == "rise" && last_val != FourValuedVector("1") && cur_val == FourValuedVector("1"))
      return true;
    if(start_trigger == "fall" && last_val != FourValuedVector("0") && cur_val == FourValuedVector("0"))
      return true;
  }
  return false;
}

void System::match_transactions() {
  vector<Transaction> to_remove;
  vector<Transaction> tmp_transactions = transactions;
  transactions.clear();
  for(vector<Transaction>::iterator it = tmp_transactions.begin();
      it != tmp_transactions.end(); ++it) {
    //          fout << "looking for matches to " << *it << endl;
    if(find(to_remove.begin(),to_remove.end(),*it) == to_remove.end()) {
      //                        fout << "not on the to_remove list\n";
      vector<Transaction>::iterator it2 = it;
      ++it2;
      for( ; it2 != tmp_transactions.end(); ++it2) {
        if(find(to_remove.begin(),to_remove.end(),*it2) == to_remove.end()) {
          if(it->matches(*it2)) {
            //                                          fout << "matched " << *it2 << endl;
            it->combine(*it2);
            to_remove.push_back(*it2);
          }
          else {
            //                                          fout << "did not match " << *it2 << endl;
          }
        }
      }
      //                        fout << "putting back " << *it << endl;
      transactions.push_back(*it);
    }
    else {
      //                        fout << "found on the to_remove list:" << endl;
      //                        for(vector<Transaction>::iterator rit = to_remove.begin();
      //                                rit != to_remove.end(); ++rit) {
      //                                mout << *rit << endl;
      //                        }
    }
  }
}

vector<int> System::find_other_transaction_ends() {
  vector<int> ends;
  vector<Transaction> tmp_transactions = transactions;
  if(tmp_transactions.size() > 1)
    tmp_transactions.pop_back();
  for(vector<Transaction>::iterator it1 = tmp_transactions.begin(); it1 != tmp_transactions.end(); ++it1) {
    //          fout << "checking outer transaction: " << *it1 << endl;
    for(vector<Transaction>::iterator it2 = tmp_transactions.begin() ; it2 != tmp_transactions.end(); ++it2) {
      if(it1 != it2) {
        //                              fout << "checking inner transaction " << *it2 << endl;
        vector<int>::const_iterator end_it = it1->find_end(*it2);
        if(end_it != it1->sequence.end() && *end_it != it1->sequence.front()) {
          mout << "found new end! " << *end_it << endl;
          mout << "from " << *it1 << " and " << *it2 << endl;
          if(find(ends.begin(),ends.end(),*(end_it)) == ends.end())
            ends.push_back(*(end_it));
        }
      }
    }
  }
  return ends;
}

void System::find_transactions() {
        
  transactions.clear();

  //    if(!transaction_mode) {
  //            print_state_diagram();
  //            print_dot_file(output_file_base + ".dot");
  //    find_legal_transitions_for_all(output_file_base + "_legal_transition_assertion.v");
  //            mout << "Finished printing dot file\n";
  //    }

  if(!transaction_mode) {
    mout << "Finding initial transaction..." << endl;
    Transaction initial_transaction = Transaction(1,state_sequence,0);
    initial_transaction.find_loops();
    fout << "-----------------------------------------\n";
    fout << "Transactions initially: " << endl;
    fout << "-----------------------------------------\n";
    fout << initial_transaction << endl;

    transactions.push_back(initial_transaction);
    return;
  }
  transactions.clear();

  find_initial_transactions(start_mode,start_signal_names,start_triggers);
  if(transactions.size() == 0)
    return;

  if(!INCREMENTAL_MODE) cout << "Finding loops.";
  for(vector<Transaction>::iterator it = transactions.begin();
      it != transactions.end(); ++it) {
    it->find_loops();
    //cout << "." << flush;
  }
  cout << endl;
  match_transactions();

  fout << "-----------------------------------------\n";
  fout << "Transactions without loops: " << endl;
  fout << "-----------------------------------------\n";
  for(vector<Transaction>::const_iterator it = transactions.begin();
      it != transactions.end(); ++it) {
    fout << *it << endl;
  }

  if(!start_mode) {
    find_other_transaction_ends();
                

    vector<int> ends = find_other_transaction_ends();
    fout << "First transaction: " << transactions.front() << endl;
    ends.push_back(transactions.front().sequence.back()); //add the original end

    fout << "Ends:\n";
    for(unsigned int i = 0; i < ends.size(); ++i) {
      fout << ends[i] << " ";
    }
    mout << endl;

    if(ends.size() > 1) {
      transactions.clear();
      vector<int> cur_transaction;
      unsigned int time_idx = 0;
      unsigned long begin_time = time_sequence[0];
      for(vector<int>::iterator it = state_sequence.begin(); it != state_sequence.end(); ++it) {
        fout << "it: " << *it << endl;
        cur_transaction.push_back(*it);
        if(find(ends.begin(),ends.end(),*it) != ends.end()) {
          fout << "found transaction\n";
          if(cur_transaction.size() > 1) {
            vector<Transaction>::iterator it = transactions.begin();
            while(it != transactions.end() && it->sequence != cur_transaction)
              ++it;
            if(it == transactions.end()) {
              transactions.push_back(Transaction(1,cur_transaction, begin_time));
            }
            else {
              ++(it->num_occurrences);
            }
            cur_transaction.clear();
          }
          ++time_idx;
          begin_time = time_sequence[time_idx];
        }
      }
      if(cur_transaction.size() > 1)
        transactions.push_back(Transaction(1,cur_transaction,begin_time));
      if(!INCREMENTAL_MODE) cout << "Finding loops.";
      for(vector<Transaction>::iterator it = transactions.begin(); it != transactions.end(); ++it) {
        it->find_loops();
        if(!INCREMENTAL_MODE)cout << "." << flush;
      } 
      cout << endl;

      match_transactions();

      fout << "-----------------------------------------\n";
      fout << "New Transactions without loops: " << endl;
      fout << "-----------------------------------------\n";
      for(vector<Transaction>::const_iterator it = transactions.begin();
          it != transactions.end(); ++it) {
        fout << *it << endl;
      }
    }
  }
  find_transaction_constants();
}

void System::find_initial_transactions(bool start_mode, const vector<string>& start_signals,
                                       const vector<string>& start_triggers) {
  //    vector<int> prefix;
  vector<int>::const_iterator it = state_sequence.begin(), last_it;
  last_it = it;
  int end=0;

  set<int> seen;
  if(!start_mode) {
    //          fout << "looking for ends, it: " << *it << endl;
    while(it != state_sequence.end() && seen.find(*it) == seen.end()) {
      //                        fout << "it: " << *it << endl;
      seen.insert(*it);
      ++it;
    }
    if(it == state_sequence.end()) {
      transactions.push_back(Transaction(1,state_sequence,time_sequence[0]));
      return;
      fout << "only one transaction occurred\n";
    }
    else {
      end = *it;
    }
  }
  //    mout << "start_mode: " << start_mode << " start_signal: " << start_signal << " start_trigger: " << start_trigger << endl;
  fout << "end:" << end << endl;
  //    mout << endl;
  vector<int> cur_transaction;
  unsigned int time_idx = 0;
  unsigned long long begin_time = time_sequence[0];
  for(vector<int>::iterator it = state_sequence.begin(); it != state_sequence.end(); ++it) {
    //          mout << "it: " << *it << endl;
    cur_transaction.push_back(*it);
    if((!start_mode && *it == end)
       || (start_mode && start_transition(*last_it,*it,start_signals,start_triggers))) {
      //                        mout << "found transaction\n";
      if(cur_transaction.size() > 1) {
        vector<Transaction>::iterator t_it = transactions.begin();
        while(t_it != transactions.end() && t_it->sequence != cur_transaction) {
          ++t_it;
        }
        if(t_it == transactions.end()) {
          transactions.push_back(Transaction(1,cur_transaction,begin_time));
        }
        else {
          ++(t_it->num_occurrences);
        }
      }
      cur_transaction.clear();
      begin_time = time_sequence[time_idx];
    }
    last_it = it;
    ++time_idx;
    //begin_time = time_sequence[time_idx];
  }
  if(cur_transaction.size() > 1)
    // don't add the transaction if it occurs at the end of the state sequence,
    // unless we are done reading the trace file, then add it anyway
    if(cur_transaction.at(cur_transaction.size() - 1) != 
       state_sequence.at(state_sequence.size() - 1) || done_parsing)
      {
        cout << "Transaction added, start time " << begin_time << endl;
        transactions.push_back(Transaction(1,cur_transaction,begin_time));
      }
}

void System::print_verilog() {
  string main_module_type = module_name_to_type_translation[main_module];
  Module m = modules[main_module_type];
  vector<string> signal_names = m.get_output_signal_names(states[0].get_signal_names());
        
  Verilog_Printer verilog_printer(states,constants,transactions,signal_names,
                                  output_file_base,transaction_mode);
  verilog_printer.found_sis = found_sis;
  verilog_printer.found_abc = found_abc;
  verilog_printer.find_legal_states();
  verilog_printer.find_legal_transitions_for_all();
}

ostream& operator<<(ostream &os, const System &s) {
  os << "-----------------------------------------\n";
  os << "Modules:\n";
  os << "-----------------------------------------\n";
  for(modules_t::const_iterator it = s.modules.begin();
      it != s.modules.end(); ++it) {
    os << it->second << endl;
  }
  os << "-----------------------------------------\n";
  os << "States: " << endl;
  os << "-----------------------------------------\n";
  for(states_t::const_iterator it = s.states.begin();
      it != s.states.end(); ++it)
    os << it->second << endl;

  os << "-----------------------------------------\n";
  os << "Constants: " << endl;
  os << "-----------------------------------------\n";
  for(constants_t::const_iterator it = s.constants.begin();
      it != s.constants.end(); ++it)
    os << it->first << " == " << it->second << endl;

  /*os << "-----------------------------------------\n";
    os << "State sequence: " << endl;
    os << "-----------------------------------------\n";
    int i = 0;
    for(vector<int>::const_iterator it = s.state_sequence.begin();
    it != s.state_sequence.end(); ++it) {
    ++i;
    os << *it << " ";
    if(i % 10 == 0)
    os << endl;
    }
    os << endl;*/

  os << "-----------------------------------------\n";
  os << "Transactions: " << endl;
  os << "-----------------------------------------\n";
  for(vector<Transaction>::const_iterator it = s.transactions.begin();
      it != s.transactions.end(); ++it) {
    os << *it << endl;
  }

  return os;
}

void System::find_correlations() {

  // look at each state
  //vector<int> state_IDs;
        
  int x;
  int y;
        
  int num_signals = 0;
        
  vector<string> signal_names;
        
  states_t::iterator s_it = states.begin();
        
  for(state_values_t::iterator it = (s_it->second).values.begin();
      it != (s_it->second).values.end(); ++it)
    {
      signal_names.push_back(it->first);
      num_signals++;
    }
        
  // matrix of results
  int corr[num_signals][num_signals];
        
  for(x=0; x<num_signals; x++)
    for(y=0; y<num_signals; y++)
      corr[x][y] = 0;
        
  for(states_t::iterator it = states.begin();it != states.end(); ++it) {
    State s = (it->second); // get state
                
    x=0;
    // examine correlations within each state
    // (would be better to look at state sequence)
    for(state_values_t::iterator v_it = s.values.begin();
        v_it != s.values.end(); ++v_it)
      {
                        
        FourValuedVector val1 = v_it->second;
                        
        y=0;
                        
        for(state_values_t::iterator v_it2 = s.values.begin();
            v_it2 != s.values.end(); ++v_it2)
          {
            FourValuedVector val2 = v_it2->second;
            //cout << "x: " << x << ", y: " << y << endl;
            //cout << val1 << " " << val2 << endl;
            if(val1 == val2) corr[x][y]++;
                                
            y++;
          }
        x++;
      }
  }
                
        
  mout << "Redundant signal pairs: " << endl;   
    
  for(x=0; x<num_signals; x++)
    {
      for(y=0; y<x; y++)
        {
          //cout << corr[x][y] << " ";
          if((corr[x][y] == corr[0][0]) || corr[x][y] == 0)
            mout << "  " << signal_names[x]
                 << ", " << signal_names[y] << endl;
        }
      //cout << endl;
    }

  mout << endl;

  for(x=0; x<num_signals; x++)
    for(y=0; y<num_signals; y++)
      corr[x][y] = 0;
        
        
  // use state sequence, rather than vector of states this time
  long transition_count = state_sequence.size() - 1;
        
  mout << "Number of state transitions examined: " << 
    transition_count << endl;
        
  vector<int>::iterator it_next = state_sequence.begin();
                
  for(vector<int>::iterator it = state_sequence.begin();
      ++it_next != state_sequence.end(); ++it) {
    State s = states[*it]; // get state
                
    State s_next = states[*it_next]; // get next state
                
    x=0;
    // examine correlations between successive states
    for(state_values_t::iterator v_it = s.values.begin();
        v_it != s.values.end(); ++v_it)
      {
        FourValuedVector val1 = v_it->second;
                        
        y=0;
                        
        for(state_values_t::iterator v_it2 = s_next.values.begin();
            v_it2 != s_next.values.end(); ++v_it2)
          {
            FourValuedVector val2 = v_it2->second;
            //cout << "x: " << x << ", y: " << y << endl;
            //cout << val1 << " " << val2 << endl;
            if(val1 == val2) corr[x][y]++;
            y++;
          }
        x++;
      }
  }
        
        
  /*for(x=0; x<num_signals; x++)
    {
    for(y=0; y<num_signals; y++)
    cout << corr[x][y] << " ";
    cout << endl;
    }*/
        
  mout << "\nHighly correlated signal pairs (over 1 cycle interval):" << endl;
        
  float corr_threshold = 0.8; // might make this a parameter
        
  for(x=0; x<num_signals; x++)
    {
      for(y=0; y<num_signals; y++)
        {
          if(corr[x][y] > corr_threshold * transition_count)
            mout << "  " << signal_names[x] << " -> " << signal_names[y] 
                 << " ("
                 << setprecision(3) << corr[x][y] / (float)transition_count
                 << ")" << endl;
          else if(corr[x][y] < (1-corr_threshold) * transition_count)
            mout << "  " << signal_names[x] << " -> ~" << signal_names[y]
                 << " ("
                 << setprecision(3) << 1 - (corr[x][y] / (float)transition_count)
                 << ")" << endl;
        }
    }
        
  mout << endl;

}

void System::clean_files() {
  string rm_cmd = "rm -f " + output_file_base + "/*.pla" + " " +
    output_file_base + "/*.eqn";;
  system(rm_cmd.c_str());
}

void System::print_state_sequence() {
  //writes entire state sequence with time (in tics) to std out
  //this must be called after entire VCD file is read

  cout << "State sequence size=" << state_sequence.size() << endl;
  cout << "Time sequence size= " << time_sequence.size() << endl;
  cout << endl << "Time             state" << endl;
  for (int i=0; i<(int)time_sequence.size(); i++) {
    cout << time_sequence[i] << " " << state_sequence[i] << endl;
  }

}

void System::print_summary() {
  // To be called at the end of a run
  // summarizes results in a few lines

  int num_states=0;
  for(states_t::const_iterator it = states.begin();
      it != states.end(); ++it)
    num_states++;

  int num_txns=0,txn_reps=0;
  unsigned long long int last_txn_time = transactions[transactions.size()-1].time_first_seen;
  for(vector<Transaction>::const_iterator it = transactions.begin();
      it != transactions.end(); ++it) {
    num_txns++;
    txn_reps += it->num_occurrences;
    if (it->time_first_seen > last_txn_time && it->num_occurrences > 1) {
      last_txn_time = it->time_first_seen;
    }
  }
  
  mout << "\n**************** Summary ********************\n";
  mout << "Name:                   " << output_file_base << "\n";
  mout << "Input File:             " << config_file_name << "\n";
  mout << "Trace File:             " << trace_file_name  << "\n";
  mout << "Number of signals:      " << signals.num_signals() 
       << " (may include clock and/or reset)\n";
  mout << "Total States:           " << num_states << "\n";
  mout << "Total Transactions:     " << num_txns << "\n";
  mout << "Total txn reps:         " << txn_reps  << "\n";
  mout << "Last repeated txn time: " << last_txn_time<< " (tics) " << "\n";
}
