/**************************************************************************
 *   Inferno : An automatic semantic information extractor                 *
 *   Copyright (C) 2006-2010  University of Michigan                       *
 *   http://www.eecs.umich.edu/inferno                                     *
 *                                                                         *
 *   Contributors include Andrew DeOrio, Adam Bauserman,                   *
 *   Valeria Bertacco and Beth Isaksen.                                    *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA         *
 *   02110-1301  USA                                                       *
 *                                                                         *
 *   Andrew DeOrio <awdeorio@umich.edu>                                    *
 *   Valeria Bertacco <valeria@umich.edu>                                  *
 *                                                                         *
 *   University of Michigan                                                *
 *   Electrical Engineering and Computer Science Dept.                     *
 *   2260 Hayward St.                                                      *
 *   Ann Arbor, MI 48109-2121                                              *
 **************************************************************************/

#include "Transaction.h"

Transaction::Transaction()
{}

Transaction::Transaction(int num_occurrences_in, const vector<int>& sequence_in, unsigned long long time_first_seen_in) : 
  num_occurrences(num_occurrences_in), sequence(sequence_in), time_first_seen(time_first_seen_in)
{
}

Transaction::~Transaction()
{}

int Transaction::size() const {
  return sequence.size();
}
void Transaction::add_state(int ID) {
  sequence.push_back(ID);
}

void Transaction::clear() {
  sequence.clear();
  loops.clear();
}

vector<int> Transaction::get_all_IDs() const {
  vector<int> IDs = sequence;
  for(vector<Loop>::const_iterator it = loops.begin();
      it != loops.end(); ++it) {
    IDs.insert(IDs.end(),it->sequence.begin(),it->sequence.end());
  }
  return IDs;
}

void Transaction::add_constant(const string& constant) {
  constants.insert(constant);
}

bool Transaction::is_constant(const string& name) const {
  return (constants.find(name) != constants.end());
}

void Transaction::expand_loop(vector<int>::iterator& loop_begin, vector<int>::iterator& loop_end,
                              const vector<int>& tmp_sequence) {
  //expand if possible
  if(loop_end != tmp_sequence.end() && loop_begin != tmp_sequence.begin()) {
    --loop_begin;
    ++loop_end;
    while(loop_begin != tmp_sequence.begin() && loop_end != tmp_sequence.end()
          && (*loop_begin) == (*loop_end)) {
      --loop_begin;
      ++loop_end;
    }
    if(loop_end == tmp_sequence.end() || *loop_begin != *loop_end) {
      ++loop_begin;
      --loop_end;
    }
  }
}

void Transaction::remove_inner_loops(Loop& cur_loop) {
  bool done = false;
  while(!done) {
    done = true;
    for(vector<Loop>::iterator past_loops_it = loops.begin();
        past_loops_it != loops.end(); ++past_loops_it) {
      if(loops.size() > 0 && cur_loop.sequence.size() > past_loops_it->sequence.size()
         && cur_loop.find_loop(*past_loops_it) != cur_loop.sequence.end()) {
        done = false;

        cur_loop.remove(*past_loops_it);
      }
      if(done)
        break;
    }
  }
}

void Transaction::find_loops() {

  //fout << "Transaction::find_loops" << endl;
  vector<int> tmp_sequence = sequence;
  Loop cur_loop;
  sequence.clear();

  int j = 0;
  vector<int>::iterator it = tmp_sequence.begin();
  while(it != tmp_sequence.end()) {
    //fout << "it: " << *it << " j: " << j << " out of " << tmp_sequence.size() << endl;
    ++j;
    sequence.push_back(*it);
    vector<int>::iterator loop_begin;
    loop_begin = find(sequence.begin(),sequence.end(),*it);
    if(loop_begin != sequence.end() - 1) { //loop found
      //fout << "found loop!" << endl;
      vector<int>::iterator peek_it = it;
      while(loop_begin != sequence.begin() && peek_it != tmp_sequence.end() && *loop_begin == *peek_it) {
        //fout << "while peek, begin: " << *loop_begin << " end: " << *peek_it << endl;
        --loop_begin;
        ++peek_it;
      }
      if(peek_it == tmp_sequence.end() || *loop_begin != *peek_it) {
        //fout << "if peek" << endl;
        ++loop_begin;
        --peek_it;
      }
                        
      cur_loop.sequence.insert(cur_loop.sequence.begin(),loop_begin,sequence.end());
      cur_loop.sequence.insert(cur_loop.sequence.end(),it+1,peek_it+1);
      it = peek_it;
      sequence.erase(loop_begin+1,sequence.end());

      if(cur_loop.sequence.size() > 1) {
        //if loop has already been found, increase the count,
        //otherwise add the new loop
        vector<Loop>::iterator loop_it = loops.begin();
        //fout << "num_loops: " << loops.size() << endl;
        int i = 0;
        while(loop_it != loops.end() && loop_it->sequence != cur_loop.sequence) {
          //                                    mout << "looking for loop, i: " << i << " out of " << loops.size() << endl;
          ++i;
          ++loop_it;
        }
        if(loop_it == loops.end()) {
          //fout << "adding new loop" << endl;
          cur_loop.min_times = 1;
          cur_loop.max_times = 1;
          loops.push_back(cur_loop);
        }
        else {
          //fout << "increasing min_times" << endl;
          ++(loop_it->min_times);
        }
      }
      cur_loop.sequence.clear();
    }
    ++it;
  }
}

//Determine whether t is contained in *this and if it is, return an iterator to
//the end
vector<int>::const_iterator Transaction::find_end(const Transaction& t) const {
  if(t.sequence.size() == 0)
    return sequence.end();
  vector<int>::const_iterator it = find(sequence.begin(),sequence.end(), t.sequence.front());
  if(it == sequence.end()) {
    return sequence.end();
  }
  vector<int>::const_iterator t_it = t.sequence.begin(), begin_it = it;
  while(it != sequence.end() && t_it != t.sequence.end() && *it == *t_it) {
    ++it;
    ++t_it;
  }
  if(t_it == t.sequence.end()) {
    for(vector<Loop>::const_iterator loop_it = t.loops.begin(); loop_it != t.loops.end(); ++loop_it) {
      if(find(loops.begin(),loops.end(),*loop_it) == loops.end())
        return sequence.end();
    }
    if(begin_it != sequence.begin())
      return begin_it - 1;
    else if(it != sequence.end())
      return it - 1;
  }
  return sequence.end();
}

bool Transaction::matches(const Transaction& t) const {
  if(sequence != t.sequence)
    return false;
  bool found_in_first;
  for(vector<Loop>::const_iterator it = loops.begin();
      it != loops.end(); ++it) {
    if(!t.has_loop(*it)) {
      found_in_first = true;
      break;
    }
  }
  if(!found_in_first)
    return true;
  for(vector<Loop>::const_iterator it = t.loops.begin();
      it != t.loops.end(); ++it) {
    if(!has_loop(*it)) {
      return false;
    }
  }
  return true;
}

bool Transaction::operator==(const Transaction& t) const {
  /*if(num_occurrences != t.num_occurrences)
    return false;*/
  if(sequence != t.sequence)
    return false;
  for(vector<Loop>::const_iterator it = loops.begin();
      it != loops.end(); ++it) {
    vector<Loop>::const_iterator it2 = find(t.loops.begin(),t.loops.end(),*it);
    if(it2 == t.loops.end())
      return false;
    if(it2->min_times != it->min_times)
      return false;
    if(it2->max_times != it->max_times)
      return false;
  }
  for(vector<Loop>::const_iterator it = t.loops.begin();
      it != t.loops.end(); ++it) {
    vector<Loop>::const_iterator it2 = find(loops.begin(),loops.end(),*it);
    if(it2 == loops.end())
      return false;
    if(it2->min_times != it->min_times)
      return false;
    if(it2->max_times != it->max_times)
      return false;
  }
  return true;
}

bool Transaction::has_loop(const Loop& l) const {
  for(vector<Loop>::const_iterator it = loops.begin(); it != loops.end(); ++it) {
    if(it->sequence == l.sequence)
      return true;
  }
  return false;
}

void Transaction::combine(const Transaction& t) {
  num_occurrences += t.num_occurrences;
  for(vector<Loop>::const_iterator it = t.loops.begin(); it != t.loops.end(); ++it) {
    for(vector<Loop>::iterator it2 = loops.begin(); it2 != loops.end(); ++it2) {
      if(it->sequence == it2->sequence) {
        if(it->max_times > it2->max_times)
          it2->max_times = it->max_times;
        if(it->min_times < it2->min_times)
          it2->min_times = it->min_times;
      }
    }
  }
  for(vector<Loop>::const_iterator it = t.loops.begin(); it != t.loops.end(); ++it) {
    if(!has_loop(*it)) {
      Loop new_loop = *it;
      new_loop.min_times = 0;
      loops.push_back(new_loop);
    }
  }
}

vector<pair<int,int> > Transaction::get_transitions() const {
  vector<pair<int,int> > transitions;

  vector<int>::const_iterator it = sequence.begin(), next_it;
  next_it = it;
  ++next_it;
  while(next_it != sequence.end()) {
    transitions.push_back(make_pair(*it,*next_it));
    ++next_it;
    ++it;
  }
  for(vector<Loop>::const_iterator loop_it = loops.begin(); loop_it != loops.end(); ++loop_it) {
    it = loop_it->sequence.begin();
    next_it = it;
    ++next_it;
    while(next_it != loop_it->sequence.end()) {
      transitions.push_back(make_pair(*it,*next_it));
      ++next_it;
      ++it;
    }
  }
  return transitions;
}

ostream& operator<<(ostream &os, const Transaction &t) {
  os << "Main transaction: " << endl;
  for(vector<int>::const_iterator it = t.sequence.begin();
      it != t.sequence.end(); ++it)
    os << *it << " ";
  os << "(occurred " << t.num_occurrences << " times)\n";
  if(t.loops.size() > 0) {
    os << "Loops:\n";
    for(vector<Loop>::const_iterator it = t.loops.begin();
        it != t.loops.end(); ++it) {
      for(vector<int>::const_iterator it2 = it->sequence.begin();
          it2 != it->sequence.end(); ++it2) {
        os << *it2 << " ";
      }
      os << "(occurred " << it->min_times << " to " << it->max_times << " times)\n";
    }
  }
  else {
    os << "(No loops.)\n";
  }
  os << "Time of first occurrence: " << t.time_first_seen << endl;
  return os;
}
