/**************************************************************************
 *   Inferno : An automatic semantic information extractor                 *
 *   Copyright (C) 2006-2010  University of Michigan                       *
 *   http://www.eecs.umich.edu/inferno                                     *
 *                                                                         *
 *   Contributors include Andrew DeOrio, Adam Bauserman,                   *
 *   Valeria Bertacco and Beth Isaksen.                                    *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA         *
 *   02110-1301  USA                                                       *
 *                                                                         *
 *   Andrew DeOrio <awdeorio@umich.edu>                                    *
 *   Valeria Bertacco <valeria@umich.edu>                                  *
 *                                                                         *
 *   University of Michigan                                                *
 *   Electrical Engineering and Computer Science Dept.                     *
 *   2260 Hayward St.                                                      *
 *   Ann Arbor, MI 48109-2121                                              *
 **************************************************************************/

#include "VCD_Parser.h"

VCD_Parser::VCD_Parser() {}
VCD_Parser::~VCD_Parser() {}

void VCD_Parser::initialize(const int max_lines_in, const int max_bits_in, const string& vcd_file_name_in,
                            const set<string>& modules_to_analyze_in,const string& clock_name_in,
                            bool signal_mode_in, bool all_signals_in,
                            const set<string>& signals_to_analyze_in, bool sync_mode_in,
                            const vector<string>& ignore_states_in) {
        
  max_lines = max_lines_in;
  max_bits = max_bits_in;
  vcd_file_name = vcd_file_name_in;
  modules_to_analyze = modules_to_analyze_in; 
  clock_name = clock_name_in;
  signal_mode = signal_mode_in;
  all_signals = all_signals_in;
  signals_to_analyze = signals_to_analyze_in;
  sync_mode = sync_mode_in;
  done_parsing = false;
  time = 0;
  total_lines = 0;
  ignore_states = ignore_states_in;
}

//Open the trace file and read the beginning to determine the translation from
//signal names to symbols
bool VCD_Parser::parseBegin() {

  vcd_file.open(vcd_file_name.c_str());
  if(!vcd_file.is_open()) {
    mout << "Failed to open trace file: " << vcd_file_name << endl;
    return false;
  }
  else {
    mout << "Opened trace file: " << vcd_file_name << endl;
    //          var_translation.clear();
    readDefinitions();
    fout << "Read definitions.\n";
  }
  return true;
}

void VCD_Parser::readDefinitions() {
  while(find_next_module_scope()) {
    read_vars();
  }
}

//return true if we actually find one
bool VCD_Parser::find_next_module_scope() {
  while(words.size() != 4 || words[0] != "$scope" || words[1] != "module") {
    //          mout << "find_next_module_scope ";
    //                  for(unsigned int i = 0; i < words.size(); ++i)
    //                          mout << words[i] << " ";
    //          mout << endl;
    if(words.size() == 4 && words[0] == "$scope") //this is a task scope -- we will ignore it
      scope.push_back(words[2]);
    if(words.size() > 0 && words[0] == "$upscope") {
      scope.pop_back();
    }
    else if(words.size() > 0 && words[0] == "$enddefinitions") {
      return false; //we're done
    }
    if(!getWords(vcd_file,words,false," \t")) {
      mout << "Error reading vcd file -- came to end without seeing $enddefinitions." << endl; //???
      return false; //also done, but with an error
    }
  }
  scope.push_back(words[2]);
  return true;
}

void VCD_Parser::read_vars() {
  string curScope = concat_string(scope,".");
  fout << "curScope: " << curScope << endl;
        
  while(getWords(vcd_file,words,false," \t") && words.size() > 0 && words[0] == "$var") {
    //          mout << "read_vars ";
    //                  for(unsigned int i = 0; i < words.size(); ++i)
    //                          mout << words[i] << " ";
    //          mout << endl;
    if(words.size() > 1 && (words[1] == "wire" || words[1] == "reg")) {
      int num_bits = atoi(words[2]);
      if(num_bits <= max_bits) {
        if(modules_to_analyze.find(curScope) != modules_to_analyze.end()) { //??? is this right?
          string full_name = curScope + "." + words[4];
          initial_var_translation.push_back(make_pair(words[3],full_name));
          initial_num_bits.push_back(num_bits);
          fout << "added translation from " << words[3] << " to " << full_name << endl;
        }
      }
    }
  }
  //    mout << "read_vars (end) ";
  //    for(unsigned int i = 0; i < words.size(); ++i)
  //            mout << words[i] << " ";
  //    mout << endl;
}

bool VCD_Parser::parseContinue(const modules_t& modules,
                               const map<string,string>& module_name_to_type_translation) {

  done_parsing = false;
  if(!vcd_file.is_open()) {
    mout << "vcd file " << vcd_file_name << " is not open.\n";
    return false;
  }
  else {
    //          fout << "File is open \n";
    done_parsing = readValues(modules, module_name_to_type_translation);
    if(done_parsing)
      vcd_file.close();
    //          fout << "Read values, done? " << done_parsing << "\n";
    fout << "number of signals read: " << data.size() << endl;
  }
  return true;
}


bool VCD_Parser::readValues(const modules_t& modules,
                            const map<string,string>& module_name_to_type_translation) {
        
  lines_of_this_chunk = 0;
  last_time_step = false;
  bool done = false;

  if(words.size() == 0)
    getWords(vcd_file, words,false," \t");
        
  while(!done) {
    if(total_lines % 100000 == 0)  //provide some occasional feedback on our progress
      cout << "." << flush;
    if(lines_of_this_chunk > max_lines) { //stop at the next reasonable place, i.e. the end of the time step
      last_time_step = true;
    }
    ++lines_of_this_chunk;
    ++total_lines;
    //          mout << "words[0]: " << words[0] << endl;
    if(!parse_line()) {
      if(!last_time_step) {
        mout << "Failed to parse line!" << endl;
        return true; //might as well stop reading
      }
      else {
        return false;
      }
    }
    getWords(vcd_file, words,false," \t");
    if(words.size() == 0)
      return true; //we're done
  }
  return false; //not done yet
}

//return true if we successfully parsed the line and should keep going
bool VCD_Parser::parse_line() {
  //    mout << "parse_line" << endl;
  if(words[0][0] == '#'){
    time = atoi(words[0].substr(1));
    //          mout << "time: " << time << endl;
    if(time < 0) {
      mout << "Negative time! " << endl; //this could arise from limitations in the size of an unsigned long
      for(unsigned int i = 0; i < words.size(); ++i)
        mout << "words[" << i << "]: " << words[i] << endl;
      return false;
    }
    if(last_time_step) {
      fout << "Last time step!" << endl;
      return false;
    }
    return true;
  }
  if(words[0][0] != '$') {
    string var_name, val;
    if(words[0][0] == 'b') {
      var_name = words[1];
      val = words[0].substr(1);
    }
    else {
      var_name = words[0].substr(1);
      val = words[0][0];
    }
    //          mout << "var_name: " << var_name << " val: " << val << endl;
    map<string,string>::const_iterator vit = var_translation.find(var_name);
    if(vit != var_translation.end()) {
      if(find(ignore_states.begin(), ignore_states.end(), val) == ignore_states.end())
        data[vit->second].values.push_back(make_pair(time,FourValuedVector(val)));
    }
    else
      {
        //mout << "Warning: var "<<var_name<<" not found!" << endl;
      } 
  }
  return true;
}

