/******************************************************************************
    Inferno : An automatic semantic information extractor
    Copyright (C) 2006-2010  University of Michigan
    http://www.eecs.umich.edu/inferno

    Contributors include Andrew DeOrio, Adam Bauserman, 
    Valeria Bertacco and Beth Isaksen.    

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301  USA

    Andrew DeOrio <awdeorio@umich.edu>
    Valeria Bertacco <valeria@umich.edu>

    University of Michigan
    Electrical Engineering and Computer Science Dept.
    2260 Hayward St.
    Ann Arbor, MI 48109-2121

******************************************************************************/

#include "Verilog_Parser.h"

Verilog_Parser::Verilog_Parser() {}

Verilog_Parser::Verilog_Parser(const vector<string>& file_names_in, const string& top_name_in):
  file_names(file_names_in), top_name(top_name_in) {}

Verilog_Parser::~Verilog_Parser() {}

bool Verilog_Parser::parse(modules_t& modules, map<string,string>& module_name_to_type_translation,
                           map<string,string>& defined_constants) {

  fout << "Parsing " << file_names.size() << " verilog files...\n";
  for(vector<string>::const_iterator it = file_names.begin(); it != file_names.end(); ++it) {
    //          mout << "Reading module definitions from verilog file: " << *it << ".\n";
    //          mout << *it << endl;
    ifstream inFile(it->c_str());       

    if(!inFile || !inFile.is_open()) {
      mout << "Unable to open verilog file " << *it << endl;
      return false;
    }
    else {
      fout << "Reading module definitions from verilog file: " << *it << ".\n";
      if(!readModuleDefs(inFile,modules,defined_constants)) {
        mout << "Failed to read module definitions.\n";
        return false;
      }
    }
    inFile.close();
  }
  if(modules.find(top_name) == modules.end()) {
    mout << "No definition encountered for the top module: \""
         << top_name << "\"\n";
    return false;
  }
  for(vector<string>::const_iterator it = file_names.begin(); it != file_names.end(); ++it) {
    fout << "Reading module instances from verilog file: " << *it << ".\n";
    //          mout << *it << endl;
    ifstream inFile(it->c_str());       

    if(!inFile || !inFile.is_open()) {
      mout << "Unable to open verilog file " << *it << endl;
      return false;
    }
    else {
      //                        mout << "File " << *it
      //                                 << " is open for reading module instances \n";
      if(!readModuleInsts(inFile,modules,module_name_to_type_translation,
                          defined_constants)) {
        mout << "Failed to read module instances.\n";
        return false;
      }
    }
    inFile.close();
  }
  //    mout << "Finished reading module instances.\n";
  return true;
}

int Verilog_Parser::parse_num(string unparsed) {
  vector<string> ops;
  vector<int> nums;

  //    mout << "unparsed: " << unparsed << endl;
  int delim_pos;
  while(unparsed.find_first_of("-+") != string::npos) {
    delim_pos = unparsed.find_first_of("-+");
    //          mout << "delim_pos: " << delim_pos << endl;
    ops.push_back(unparsed.substr(delim_pos, 1));
    nums.push_back(atoi(unparsed.substr(0,delim_pos)));
    unparsed = unparsed.substr(delim_pos+1);
    //          mout << "nums: " << nums.back() << " ops: " << ops.back() << endl;
    //          mout << "unparsed: " << unparsed << endl;
  }
  //    mout << "last delim_pos: " << delim_pos << endl;
  nums.push_back(atoi(unparsed));
  int parsed = nums[0];
  for(unsigned int i = 0; i < ops.size(); ++i) {
    //          mout << "ops[" << i << "]: " << ops[i] << " nums[" << i+1 << "]: " << nums[i+1] << endl;
    if(ops[i] == "-") {
      parsed -= nums[i+1];
    }
    else {
      parsed += nums[i+1];
    }
  }
  //    mout << "parsed: " << parsed << endl;
  return parsed;
}

int Verilog_Parser::get_num_bits(string range) {
  //    mout << "range: " << range << endl;
  int colon_pos = range.find_first_of(':');
  string first_string = range.substr(1,colon_pos-1);
  string second_string = range.substr(colon_pos+1,range.size() - colon_pos - 2);

  int first = parse_num(first_string);
  int second = parse_num(second_string);

  //    mout << "num_bits: " << max(first+1,second+1) << endl;
  return max(first+1,second+1);
}


void Verilog_Parser::add_signal_to_module(const vector<string>& words, Module& m) {
  //  mout << "begin add_signal_to_module\n";
        
  unsigned int i = 1;
  int num_bits=0;
  string signal_name;

  //    mout << "add signal to module\n";
  while(i < words.size()) {
    //          mout << "words[" << i << "] = " << words[i] << endl;
    if(words[i][0] != '[') {
      //                        mout << "not bit range\n";
      signal_name = words[i];
      num_bits = 1;
    }
    else {
      //                        mout << "bit range\n";
      if(i+1 < words.size()) {
        signal_name = words[i+1];
        num_bits = get_num_bits(words[i]);
        //                              signal = make_pair(get_num_bits(words[i]),words[i+1]);
        ++i;
      }
      else {
        //                              mout << "words[i]: " << words[i] << " words[i+1]: " << words[i+1] << endl;
        mout << "Error in verilog near " << words[i] << "\n";
        //                              return;
      }
    }
    //          mout << "signal: " << signal.first << " " << signal.second << endl;
    if(words[0] == "input") {
      m.inputs[signal_name] = num_bits;
    }
    else if(words[0] == "output" || words[0] == "inout") {
      m.outputs[signal_name] = num_bits;
    }
    else if(words[0] == "reg" || words[0] == "wire" || words[0] == "integer") {
      if(m.inputs.find(signal_name) == m.inputs.end()
         && m.outputs.find(signal_name) == m.outputs.end())
        m.internals[signal_name] = num_bits;
      //                        if(find(m.inputs.begin(),m.inputs.end(),signal) == m.inputs.end()
      //                           && find(m.outputs.begin(),m.outputs.end(),signal) == m.outputs.end())
      //                                m.internals.push_back(signal);
    }
    ++i;
  }
}

bool Verilog_Parser::readModuleDefs(ifstream& inFile, modules_t& modules,
                                    map<string,string>& defined_constants) {

  Module cur_module;
  vector<string> words;
  bool in_task = false;
  //    unsigned int ifdef = 0, elsdef = 0;
  map<string,string> parameters;

  getWords(inFile,words,true,defined_constants);
  while(words.size() > 0) {
    //                  for(unsigned int i = 0; i < words.size(); ++i) {
    //                          mout << "words[" << i << "] " << words[i] << " ";
    //                  }
    //                  mout << endl;
    if(in_task) {
      if(words[0] == "endtask") {
        in_task = false;
      }
    }
    if(words[0] == "module") {
      cur_module.type = words[1];
    }
    else if(words[0] == "input" || words[0] == "output" || words[0] == "inout"
            || words[0] == "reg" || words[0] == "wire" || words[0] == "integer") {
      for(unsigned int i = 1; i < words.size(); ++i) {
        for(map<string,string>::iterator p_it = parameters.begin();
            p_it != parameters.end(); ++p_it) {
          if(words[i].find(p_it->first) != string::npos) {
            //                                                  mout << "word " << words[i] << " matched parameter: " << p_it->first << endl;
            string tmp0, tmp1;
            unsigned int start = words[i].find(p_it->first);
            unsigned int len = p_it->first.size();
            tmp0 = words[i].substr(0,start);
            tmp1 = words[i].substr(start + len);
            words[i] = tmp0 + p_it->second + tmp1;
            //                                                  mout << "changed to " << words[i] << endl;
          }
        }
      }
      add_signal_to_module(words, cur_module);
    }
    else if(words[0] == "endmodule") {
      modules[cur_module.type] = cur_module;
      fout << "added module def: " << cur_module.type << endl;
      cur_module.clear();
    }
    else if(words[0] == "task") {
      in_task = true;
    }
    else if(words[0] == "parameter") {
      if(words.size() >= 4) {
        parameters[words[1]] = words[3];
        //                              mout << "added parameter " << words[1] << " == " << words[3] << endl;
      }
    }
    getWords(inFile,words,true,defined_constants);
  }
  return true;
}

bool Verilog_Parser::readModuleInsts(ifstream& inFile, modules_t& modules, map<string,string>& module_name_to_type_translation,
                                     map<string,string>& defined_constants) {
  vector<string> words;
  string cur_line;
  string module_string;
  string parent;
  Module_Child cur_inst;

  //    mout << "readModuleInsts\n";
  while(getLineNoComments(inFile,cur_line,defined_constants)) {
    //          mout << "cur_line: " << cur_line << endl;
    cur_line = cur_line.substr(0,cur_line.find_first_of("//")); //strip comments
    if(splitString(cur_line," \t(),;",words) > 0) {
      if(words[0] == "module") {
        parent = words[1];
      }
      else if(modules.find(words[0]) != modules.end()) {
        //read in the whole instance
        module_string = "";
        while(cur_line.find_first_of(';') == string::npos) {
          //                                    mout << "in module, cur_line: " << cur_line << endl;
          module_string += cur_line;
          if(!getLineNoComments(inFile,cur_line,defined_constants)) {
            mout << "Incomplete module encountered.\n";
            return false;
          }
          //                                    mout << "module_string: " << module_string << endl;
        }
        module_string += cur_line;
        //skip parameters
        if(module_string.find('#') != string::npos) {
          int first_open_paren = module_string.find_first_of('(');
          int first_close_paren = module_string.find_first_of(')');
          module_string.erase(first_open_paren,
                              first_close_paren - first_open_paren + 1);
        }
        //re-split
        int num_words = splitString(module_string," \t,;#",words);
        cur_inst.parent = parent;
        if(num_words < 2) {
          mout << "Error: Module instance must have at least two words.\n";
          return false;
        }
        //                              for(unsigned int i = 0; i < words.size(); ++i)
        //                                      mout << "words[" << i << "]: " << words[i] << endl;
        cur_inst.type = words[0];
        cur_inst.name = words[1];
        string::size_type pos = cur_inst.name.find_first_of("(");
        if(pos != string::npos) {
          cur_inst.name = cur_inst.name.substr(0,pos);
        }
        string::size_type open_paren_pos = module_string.find_first_of("(");
        string::size_type close_paren_pos = module_string.find_last_of(")"); 
        module_string = module_string.substr(open_paren_pos + 1,close_paren_pos - open_paren_pos - 1);
        string::size_type first_non_space_pos = module_string.find_first_not_of(" \t");
        string::size_type last_non_space_pos = module_string.find_last_not_of(" \t");
        module_string = module_string.substr(first_non_space_pos, last_non_space_pos - first_non_space_pos + 1);
        //                              mout << "module_string: " << module_string << endl;
        bool done = false;
        string argFormal, argActual;
        bool found_pair;
        while(!done) {
          found_pair = false;
          //                                    mout << "start while" << endl;
          string::size_type dot_pos = module_string.find_first_of(".");
          assert(dot_pos != string::npos);
          string::size_type open_paren_pos = module_string.find_first_of("("); 
          assert(open_paren_pos != string::npos);
          string::size_type close_paren_pos = module_string.find_first_of(")");
          assert(close_paren_pos != string::npos);
          //                                    mout << "dot_pos: " << dot_pos << " open_paren_pos: " << open_paren_pos
          //                                             << " close_paren_pos: " << close_paren_pos << endl;
          if(close_paren_pos > open_paren_pos + 1) {
            argFormal = module_string.substr(dot_pos + 1,open_paren_pos - dot_pos - 1);
            argActual = module_string.substr(open_paren_pos+1,close_paren_pos - open_paren_pos - 1);
            string::size_type first_non_space_pos = argFormal.find_first_not_of(" \t");
            assert(first_non_space_pos != string::npos);
            string::size_type last_non_space_pos = argFormal.find_last_not_of(" \t");
            assert(last_non_space_pos != string::npos);

            // if the formal arg is not empty, trim whitespace
            if(last_non_space_pos - first_non_space_pos > 0) {
              argFormal = argFormal.substr(first_non_space_pos, last_non_space_pos - first_non_space_pos + 1);
              // now trim whitespace from actual arg (in parentheses)
              first_non_space_pos = argActual.find_first_not_of(" \t");
              last_non_space_pos = argActual.find_last_not_of(" \t");
              // this arg may be empty
              if(last_non_space_pos - first_non_space_pos > 0) {
                argActual = argActual.substr(first_non_space_pos, last_non_space_pos - first_non_space_pos + 1);
                //                              mout << "argFormal: " << argFormal << " argActual: " << argActual << endl;
                found_pair = true;
              }
            }
          }
          module_string = module_string.substr(close_paren_pos+1);
          //                                    mout << "module_string: \'" << module_string << "\'" << endl;
          if(module_string.size() == 0)
            done = true;
          else if(found_pair)
            cur_inst.connections.push_back(make_pair(argFormal,argActual));
          //                                    mout << "end while" << endl;
        }                               
        modules[parent].children[cur_inst.name] = cur_inst;
        module_name_to_type_translation[parent + "." + cur_inst.name] = cur_inst.type;
        fout << "Added name to type translation from " << parent + "." + cur_inst.name << " to " << cur_inst.type << endl;
        cur_inst.clear();
      }
    }
    words.clear();
  }
  return true;
}
