/******************************************************************************
    Inferno : An automatic semantic information extractor
    Copyright (C) 2006-2010  University of Michigan
    http://www.eecs.umich.edu/inferno

    Contributors include Andrew DeOrio, Adam Bauserman, 
    Valeria Bertacco and Beth Isaksen.    

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301  USA

    Andrew DeOrio <awdeorio@umich.edu>
    Valeria Bertacco <valeria@umich.edu>

    University of Michigan
    Electrical Engineering and Computer Science Dept.
    2260 Hayward St.
    Ann Arbor, MI 48109-2121

******************************************************************************/

#include "Verilog_Printer.h"

Verilog_Printer::Verilog_Printer() {}

Verilog_Printer::Verilog_Printer(const map<int,State>& states_in, const map<string,FourValuedVector>& constants_in,
                                 const vector<Transaction>& transactions_in,
                                 const vector<string>& signal_names_in, const string& output_file_base_in,
                                 bool transaction_mode_in)
  : states(states_in), constants(constants_in), transactions(transactions_in), signal_names(signal_names_in),
    output_file_base(output_file_base_in), transaction_mode(transaction_mode_in) {

  for(map<string,FourValuedVector>::iterator it = constants.begin(); it != constants.end(); ++it)
    constant_names.push_back(it->first);
}

Verilog_Printer::~Verilog_Printer() {}


// call tool to do logic minimization (SIS or ABC)
void Verilog_Printer::minimize_logic(string input_file, string output_file) {
    
  if(found_sis)
    {
      string cmd = string("echo -e \"sweep; collapse\nfull_simplify -m nocomp\nwrite_pla;\nsweep;\nwrite_pla\ndecomp\nwrite_eqn\"") +
        "| sis -t pla -T eqn -o " + output_file + " < " + input_file + " > /dev/null";
      system(cmd.c_str());
    }
  else if(found_abc)
    {
      string cmd = "echo \"read_pla " + input_file + "; balance; rewrite; refactor; balance;" +
        " rewrite; rewrite -z; balance; refactor -z; refactor -z; balance; write_eqn " + 
        output_file + "\" | abc > /dev/null";
      system(cmd.c_str());
    }
}


void Verilog_Printer::find_legal_states() {
  if(states.size() < 2)
    return;
        
  ofstream out_file;
  string f_name = output_file_base + "state.pla";
  out_file.open(f_name.c_str(),ofstream::out);
  if(!out_file.is_open()) {
    mout << "Unable to open file " << f_name << " for printing pla file.\n";
    return;
  }
  else {
    fout << "Opened file " << f_name << " for printing pla file.\n";
  }

  out_file << "# " + output_file_base + " legal states\n";
  out_file << ".i " << signal_names.size() << endl;
  out_file << ".o 1" << endl;
  out_file << ".ilb " << square_to_pointy_braces(get_short_names_as_string(signal_names)) << endl;
  //    for(vector<string>::iterator it = signal_names.begin(); it != signal_names.end(); ++it)
  //            out_file << it->substr(it->find_last_of(".") + 1) << " ";
  //    out_file << endl;
  out_file << ".ob legal_state" << endl;
  if(found_sis) out_file << ".type f" << endl;
  for(map<int,State>::iterator it = states.begin(); it != states.end(); ++it) {
    if(it == states.begin())
      ++it;
    out_file << it->second.get_signal_values_string(signal_names);
    out_file << " 1" << endl;
  }
  out_file << ".e" << endl;

  out_file.close();

  minimize_logic(output_file_base + "state.pla", output_file_base       + "state.eqn");
}

void Verilog_Printer::print_legal_state_checker_verilog() {

  vector<string> inputs, outputs, assigns, wires;
  string uniquify_string = "";
  string eqn_file_name = output_file_base + "state.eqn";
        
  assigns_from_eqn(eqn_file_name,uniquify_string,
                   inputs, outputs, assigns, wires);
        
  string out_file_name = output_file_base + "state_assertion.v";
  ofstream out_file(out_file_name.c_str());
  if(!out_file.is_open()) {
    mout << "Failed to open output file: " << out_file_name << endl;
    return;
  }
  out_file << "module legal_state_assertion(reset, " << format_vector("%s, ",inputs)
           << format_vector("%s",outputs,", ") 
           << format_vector("%s",constant_names,", ") << ");" << endl;
  out_file << "   input reset;" << endl;
  out_file << format_inputs(inputs) << endl;
  out_file << "   output fail;" << endl << endl;
  out_file << format_wires(wires) << endl;
        
  if(constants.size() > 0) {
    out_file << endl;
    out_file << "   wire legal_constants;" << endl;
  }
  out_file << endl;
  out_file << "   assign fail = ~" << outputs[0];
  if(constants.size() > 0)
    out_file << " & legal_constants";
  out_file << ";" << endl << endl;


  if(constants.size() > 0) {
    out_file << "   assign legal_constants = ";
    for(map<string,FourValuedVector>::iterator it = constants.begin(); it != constants.end(); ++it) {
      if(it != constants.begin())
        out_file << " & ";
      out_file << "(" << get_short_name(it->first) << " == 1'b" << it->second << ")";
    }
    out_file << ";" << endl;
    out_file << endl;
  }

  out_file << format_assigns(assigns) << endl;
  out_file << get_checker_warning();
  out_file << "endmodule" << endl;
}

void Verilog_Printer::find_legal_transitions_by_signal(const string& f_name_base,const string& signal_name) {

  vector<string> other_signal_names = signal_names;
  vector<string>::iterator it = find(other_signal_names.begin(), other_signal_names.end(), signal_name);
  other_signal_names.erase(it);
  vector<string> signal_name_vector;
  signal_name_vector.push_back(signal_name);
        
  ofstream out_file;

  string pla_f_name = f_name_base + ".pla";
  string eqn_f_name = f_name_base + ".eqn";

  out_file.open(pla_f_name.c_str(),ofstream::out);
  if(!out_file.is_open()) {
    mout << "Unable to open file " << pla_f_name << " for printing pla file.\n";
    return;
  }
  else {
    fout << "Opened file " << pla_f_name << " for printing pla file.\n";
  }

  out_file << "# " + output_file_base + " legal transitions by signal\n";
  out_file << ".i " << 2*(signal_names.size()) - 1<< endl;
  out_file << ".o 1" << endl;
        
  out_file << ".ilb ";

  out_file << square_to_pointy_braces(get_short_names_as_string(signal_names,"","_last")) << " ";
  out_file << square_to_pointy_braces(get_short_names_as_string(other_signal_names));
        
  out_file << endl;
  out_file << ".ob " << square_to_pointy_braces(get_short_name(signal_name)) << endl;
  if(found_sis) out_file << ".type fr" << endl;
        
  State from_state, to_state;
  for(map<int,State>::iterator it = states.begin(); it != states.end(); ++it) {

    from_state = it->second;
    vector< pair<int,long> > nexts = it->second.nexts;
                
    for(vector< pair<int,long> >::iterator next_it = nexts.begin();
        next_it != nexts.end(); ++next_it) {
                        
      to_state = states[next_it->first];

      if(to_state.get_signal_values_string(signal_name_vector) != "-") {
        out_file << from_state.get_signal_values_string(signal_names);
        out_file << to_state.get_signal_values_string(other_signal_names);
        out_file << " ";
        out_file << to_state.get_signal_values_string(signal_name_vector);
        out_file << endl;
      }
    }
  }
  out_file << ".e" << endl;

  out_file.close();

  minimize_logic(pla_f_name, eqn_f_name);
}

void Verilog_Printer::find_legal_transitions_for_transaction(Transaction t, string f_name_base) {
  ofstream out_file;

  string pla_f_name = f_name_base + ".pla";
  string eqn_f_name = f_name_base + ".eqn";

  out_file.open(pla_f_name.c_str(),ofstream::out);
  if(!out_file.is_open()) {
    mout << "Unable to open file " << pla_f_name << " for printing pla file.\n";
    return;
  }
  else {
    fout << "Opened file " << pla_f_name << " for printing pla file.\n";
  }

  out_file << "# " + output_file_base + " legal transitions for transaction\n";

  out_file << ".i ";
  if(transaction_mode) {
    out_file << 2*(signal_names.size()) + 1;
  }
  else {
    out_file << 2*(signal_names.size());
  }
  out_file << endl;
  out_file << ".o 1" << endl;

  out_file << ".ilb ";
  if(transaction_mode)
    out_file << "legal_end_last ";
  out_file << square_to_pointy_braces(get_short_names_as_string(signal_names,"","_last")) << " ";
  out_file << square_to_pointy_braces(get_short_names_as_string(signal_names)) << endl;

  out_file << ".ob legal_transitions" << endl;
  if(found_sis)out_file << ".type f" << endl;
        
  vector<pair<int,int> > transitions = t.get_transitions();
  State from_state, to_state;
  for(vector<pair<int,int> >::iterator it = transitions.begin(); it != transitions.end(); ++it) {
    //          mout << "transition from: " << it->first << " to: " << it->second << endl;

    from_state = states[it->first];
    to_state = states[it->second];

    if(transaction_mode && it == transitions.begin()) {
      out_file << "1";  //legal_end_last == 1
      for(unsigned int i = 0; i < signal_names.size(); ++i)
        out_file << "-";
      out_file << from_state.get_signal_values_string(signal_names);
      out_file << " 1" << endl;
    }

    if(transaction_mode)
      out_file << "0"; //legal_end_last == 0
    out_file << from_state.get_signal_values_string(signal_names);
    out_file << to_state.get_signal_values_string(signal_names);

    out_file << " 1" << endl;
  }
  out_file << ".e" << endl;

  out_file.close();

  minimize_logic(pla_f_name, eqn_f_name);
}

void Verilog_Printer::find_legal_ends(const set<int>& ends, string f_name_base) {
  ofstream out_file;

  string pla_f_name = f_name_base + ".pla";
  string eqn_f_name = f_name_base + ".eqn";

  out_file.open(pla_f_name.c_str(),ofstream::out);
  if(!out_file.is_open()) {
    mout << "Unable to open file " << pla_f_name << " for printing pla file.\n";
    return;
  }
  else {
    fout << "Opened file " << pla_f_name << " for printing pla file.\n";
  }

  int num_inputs = states[0].get_num_signal_bits(signal_names);

  out_file << "# " + output_file_base + " legal ends\n";
  out_file << ".i " << num_inputs << endl;
  out_file << ".o 1" << endl;
  out_file << ".ilb " << square_to_pointy_braces(get_short_names_as_string(signal_names,"","_last")) << endl;
  out_file << ".ob legal_end_last" << endl;
  if(found_sis) out_file << ".type f" << endl;
  State cur_state;
  for(set<int>::iterator it = ends.begin(); it != ends.end(); ++it) {

    cur_state = states[*it];
    out_file << cur_state.get_signal_values_string(signal_names);
    out_file << " 1" << endl;

  }
  out_file << ".e" << endl;

  out_file.close();

  minimize_logic(pla_f_name, eqn_f_name);
}


void Verilog_Printer::find_legal_transitions_for_all() {

  if(transactions.size() == 0) {
    mout << "Error: cannot find legal transitions -- transactions.size() == 0" << endl;
    return;
  }
        
  if(transaction_mode) {
    unsigned long long int i = 0;
    string f_name_base = output_file_base + "transition_";
    set<int> ends;
    for(vector<Transaction>::iterator it = transactions.begin(); it != transactions.end(); ++it) {
      find_legal_transitions_for_transaction(*it,f_name_base + itoa(i));
      ++i;
      ends.insert(it->sequence.back());
    }
    find_legal_ends(ends,output_file_base + "ends");
  }
  else {
    string f_name_base = output_file_base + "transition";
    find_legal_transitions_for_transaction(transactions.front(),f_name_base);
  }


  find_legal_states();
  print_legal_state_checker_verilog();

  string f_name_base = output_file_base + "transition_by_signal_";
  set<int> ends;
  for(vector<string>::iterator it = signal_names.begin(); it != signal_names.end(); ++it) {
    find_legal_transitions_by_signal(f_name_base + get_really_short_name(*it),*it);
    print_legal_transitions_by_signal(f_name_base + get_really_short_name(*it),*it);
  }
        
  //mout << "system::find_legal_transitions_for_all" << endl;
  string out_file_name = output_file_base + "transition_assertion.v";
  if(transaction_mode)
    print_legal_transition_all_verilog(out_file_name);
  else
    print_legal_transition_verilog(out_file_name);

}

void Verilog_Printer::print_legal_transitions_by_signal(const string& f_name_base, const string& signal_name) {
  vector<string> inputs, outputs, assigns, wires;

  //mout << "print verilog by signal" << endl;
  string eqn_file_name = f_name_base + ".eqn";
  //mout << "eqn_file_name: " << eqn_file_name << endl;
  assigns_from_eqn(eqn_file_name,"",inputs, outputs, assigns, wires);

  string out_file_name = f_name_base + ".v";
  ofstream out_file(out_file_name.c_str());
  if(!out_file.is_open()) {
    mout << "Failed to open output file: " << out_file_name << endl;
    return;
  }

  vector<string> short_signal_names = get_short_names(signal_names);
  vector<string> short_constant_names = get_short_names(constant_names);
  vector<string> signals_and_constants = short_signal_names;
  signals_and_constants.insert(signals_and_constants.end(), short_constant_names.begin(), short_constant_names.end());

  out_file << "module legal_transition" << get_really_short_name(signal_name)
           << "_assertion(reset, " << format_vector("%s, ",signals_and_constants) << "fail);" << endl;
  out_file << "   input reset;" << endl;
  out_file << format_inputs(signals_and_constants);
  out_file << endl;
  out_file << "   output fail;" << endl;
  out_file << endl;
  out_file << format_regs(signals_and_constants,"_reg");
  out_file << endl;
  out_file << format_regs(signals_and_constants,"_last");
  out_file << endl;
  out_file << "   wire legal_transition;" << endl;

  wires.erase(wires.begin());
  out_file << format_wires(wires);
        
  if(constants.size() > 0) {
    out_file << endl << "   wire legal_constants;" << endl;
  }
  out_file << endl;
  if(constants.size() > 0)
    out_file << "   assign fail = ~(legal_transition & legal_constants);" << endl;
  else
    out_file << "   assign fail = ~legal_transition;" << endl;
  out_file << endl;
        
  string first_assign = assigns.front();
  assigns.erase(assigns.begin());
  string lhs = first_assign.substr(0,first_assign.find("=") - 1);
  string rhs = first_assign.substr(first_assign.find("=") + 2);
  out_file << "   assign legal_transition = (" << lhs << " == (" << rhs << "));" << endl;
  out_file << format_assigns(assigns) << endl;

  if(constants.size() > 0) {
    out_file << "   assign legal_constants = ";
    for(map<string,FourValuedVector>::iterator it = constants.begin(); it != constants.end(); ++it) {
      if(it != constants.begin())
        out_file << " & ";
      out_file << "(" << get_short_name(it->first) << " == 1'b" << it->second << ")";
    }
    out_file << ";" << endl;
    out_file << endl;
  }

  out_file << format_assigns(assigns) << endl;
  out_file << "   always @(*)" << endl; 
  //    out_file << "   always @(" << format_vector("%s or ",signals_and_constants) << "reset) " << endl;
  out_file << "     begin" << endl;
  out_file << endl;
  out_file << format_vector("      %s_reg <= (~reset & %s);\n",signals_and_constants);
  out_file << endl;
  out_file << format_vector("      %s_last <= (~reset & %s_reg);\n",signals_and_constants);
  out_file << "   end // always @" << endl;
  out_file << endl;

  out_file << "//synopsys translate_off" << endl;
  out_file << "   always @(*)" << endl; //" << format_vector("%s or ",signals_and_constants) << "reset) " << endl;
  out_file << "     begin" << endl;
  out_file << endl;
  out_file << "     casex(legal_transition)" << endl;
  out_file << "       1'b0 : legal_transition = 1'b0;" << endl;
  out_file << "       1'b1 : legal_transition = 1'b1;" << endl;
  out_file << "     endcase" << endl;
  out_file << endl;
  out_file << "     end //always" << endl;
  out_file << "//synopsys translate_on" << endl;

  out_file << get_checker_warning();
  out_file << "endmodule" << endl;
        
}

void Verilog_Printer::print_legal_transition_verilog(const string& out_file_name) {
        
  vector<string> inputs, outputs, assigns, wires, cur_assigns, cur_wires;
  string eqn_file_name = output_file_base + "transition.eqn";

  assigns_from_eqn(eqn_file_name,"",
                   inputs, outputs, cur_assigns, cur_wires);
  assigns.insert(assigns.end(),cur_assigns.begin(),cur_assigns.end());
  wires.insert(wires.end(),cur_wires.begin(),cur_wires.end());

  ofstream out_file(out_file_name.c_str());
  if(!out_file.is_open()) {
    mout << "Failed to open output file: " << out_file_name << endl;
    return;
  }

  vector<string> short_signal_names = get_short_names(signal_names);
  vector<string> short_constant_names = get_short_names(constant_names);
        
  out_file << "module legal_transition_assertion(reset, " << format_vector("%s, ",short_signal_names)
           << format_vector("%s, ",short_constant_names) << "fail);" << endl;
  out_file << "   input reset;" << endl;
  out_file << format_inputs(short_signal_names);
  out_file << format_inputs(short_constant_names);
  out_file << endl;
  out_file << "   output fail;" << endl;
  out_file << endl;
  out_file << format_regs(short_signal_names,"_reg");
  out_file << format_regs(short_constant_names,"_reg");
  out_file << endl;
  out_file << format_regs(short_signal_names,"_last");
  out_file << format_regs(short_constant_names,"_last");
  out_file << endl;
  out_file << "   reg legal_transition;" << endl;
        
  if(constants.size() > 0) {
    out_file << endl;
    out_file << "   wire legal_constants;" << endl;
  }
  out_file << endl;
  out_file << "   assign fail = ~legal_transition";
  if(constants.size() > 0)
    out_file << " & legal_constants";
  out_file << ";" << endl << endl;


  if(constants.size() > 0) {
    out_file << "   assign legal_constants = ";
    for(map<string,FourValuedVector>::iterator it = constants.begin(); it != constants.end(); ++it) {
      if(it != constants.begin())
        out_file << " & ";
      out_file << "(" << get_short_name(it->first) << " == 1'b" << it->second << ")";
    }
    out_file << ";" << endl;
    out_file << endl;
  }

  out_file << format_assigns(assigns) << endl;
  out_file << endl;
  out_file << "   always @(*)" << endl;
  out_file << "     begin" << endl;
  out_file << endl;
  for(vector<string>::const_iterator it = signal_names.begin(); it != signal_names.end(); ++it) {
    out_file << "      " << *it << "_reg <= (~reset & " << *it << ");" << endl;
  }
  for(vector<string>::const_iterator it = constant_names.begin(); it != constant_names.end(); ++it) {
    out_file << "      " << *it << "_reg <= (~reset & " << *it << ");" << endl;
  }
  out_file << endl;
  for(vector<string>::const_iterator it = signal_names.begin(); it != signal_names.end(); ++it) {
    out_file << "      " << *it << "_last <= (~reset & " << *it << "_reg);" << endl;
  }
  for(vector<string>::const_iterator it = constant_names.begin(); it != constant_names.end(); ++it) {
    out_file << "      " << *it << "_last <= (~reset & " << *it << "_reg);" << endl;
  }
  out_file << "   end // always @";

  out_file << endl;
  out_file << "//synopsys translate_off" << endl;
  //    out_file << "   always @(" << format_vector("%s or ",short_signal_names)
  //                     << format_vector("%s or ",short_constant_names)
  //                     << " reset) " << endl;
  out_file << "   always @(*)" << endl;
  out_file << "     begin" << endl;
  out_file << endl;
  out_file << "     casex(legal_transitions)" << endl;
  out_file << "       1'b0 : legal_transition = 1'b0;" << endl;
  out_file << "       1'b1 : legal_transition = 1'b1;" << endl;
  out_file << "     endcase" << endl;
  out_file << endl;
  out_file << "     end //always" << endl;
  out_file << "//synopsys translate_on" << endl;
  out_file << endl;
  out_file << get_checker_warning();
  out_file << "endmodule" << endl;
}

void Verilog_Printer::print_legal_transition_all_verilog(const string& out_file_name) {
  vector<string> inputs, outputs, assigns, wires, cur_assigns, cur_wires;

  for(unsigned int i = 0; i < transactions.size(); ++i) {
    char index[512];
    sprintf(index,"%d",i);
    string eqn_file_name = output_file_base + "transition_" + string(index) + ".eqn";
    string uniquify_string = "transaction_" + string(index);
    assigns_from_eqn(eqn_file_name,uniquify_string,
                     inputs, outputs, cur_assigns, cur_wires);
    assigns.insert(assigns.end(),cur_assigns.begin(),cur_assigns.end());
    wires.insert(wires.end(),cur_wires.begin(),cur_wires.end());
    cur_assigns.clear();
    cur_wires.clear();
  }
  string eqn_file_name = output_file_base + "ends.eqn";
  string uniquify_string = "";
  assigns_from_eqn(eqn_file_name,uniquify_string,
                   inputs, outputs, cur_assigns, cur_wires);
  assigns.insert(assigns.end(),cur_assigns.begin(),cur_assigns.end());
  wires.insert(wires.end(),cur_wires.begin(),cur_wires.end());
        
  //    string out_file_name = output_file_base + "transition_assertion.v";
  ofstream out_file(out_file_name.c_str());
  if(!out_file.is_open()) {
    mout << "Failed to open output file: " << out_file_name << endl;
    return;
  }

  vector<string> short_signal_names = get_short_names(signal_names);
  vector<string> short_constant_names = get_short_names(constant_names);
  vector<string> signals_and_constants = short_signal_names;
  signals_and_constants.insert(signals_and_constants.end(),short_constant_names.begin(),short_constant_names.end());

  out_file << "module legal_transition_assertion(clock, reset, " << format_vector("%s, ",signals_and_constants)
           << "fail);" << endl << endl;
  out_file << "   input reset;" << endl;
  out_file << format_inputs(signals_and_constants) << endl;
  out_file << "   output fail;" << endl;
  out_file << endl;
  out_file << format_regs(signals_and_constants,"_reg") << endl;
  out_file << format_regs(signals_and_constants,"_last") << endl;
  out_file << "   reg [" << (transactions.size() - 1) << ":0] legal_transaction;" << endl;
  out_file << "   reg [" << (transactions.size() - 1) << ":0] legal_transaction_last;" << endl;
  out_file << endl;
  out_file << "   wire legal_transition;" << endl;
  out_file << endl;
  out_file << format_wires(wires)
           << endl;
  if(constants.size() > 0) {
    out_file << endl;
    out_file << "   wire legal_constants;" << endl;
  }
  out_file << endl;
  out_file << "   assign fail = ~legal_transition";
  if(constants.size() > 0)
    out_file << " & legal_constants";
  out_file << ";" << endl << endl;


  if(constants.size() > 0) {
    out_file << "   assign legal_constants = ";
    for(map<string,FourValuedVector>::iterator it = constants.begin(); it != constants.end(); ++it) {
      if(it != constants.begin())
        out_file << " & ";
      out_file << "(" << get_short_name(it->first) << " == 1'b" << it->second << ")";
    }
    out_file << ";" << endl;
    out_file << endl;
  }
        
  //    out_file << "   assign fail = ~legal_transition;" << endl << endl;
  out_file << "   assign legal_transition =   "
           << make_vector_string("legal_transaction_last[]","","","","\n\t\t\t     | ",0,transactions.size() - 1);
  out_file << ";" << endl;
  out_file << endl;
  out_file << format_assigns(assigns);
  out_file << endl;

  out_file << "   always @(posedge clock)" << endl;
  out_file << "     begin" << endl;
  out_file << endl;
  for(vector<string>::const_iterator it = signals_and_constants.begin(); it != signals_and_constants.end(); ++it) {
    out_file << "      " << *it << "_reg <= (~reset & " << *it << ");" << endl;
  }
  out_file << endl;
  for(vector<string>::const_iterator it = signals_and_constants.begin(); it != signals_and_constants.end(); ++it) {
    out_file << "      " << *it << "_last <= (~reset & " << *it << "_reg);" << endl;
  }
  out_file << endl;
  out_file << make_vector_string("         legal_transaction_last[] <= (~reset & legal_transaction[] & (legal_end_last | legal_transaction_last[]));\n","","","","",0,transactions.size() - 1);
  out_file << endl;
  out_file << "     end //else" << endl;
  out_file << "   end //always" << endl;
  out_file << endl;
  out_file << "//synopsys translate_off" << endl;
  out_file << "   always @(posedge clock)" << endl;
  out_file << "     begin" << endl;
  out_file << endl;
  for(unsigned int i = 0; i < transactions.size(); ++i) {
    out_file << "     casex(transaction_" << i << "transitions)" << endl;
    out_file << "       1'b0 : legal_transaction[" << i << "] = 1'b0;" << endl;
    out_file << "       1'b1 : legal_transaction[" << i << "] = 1'b1;" << endl;
    out_file << "     endcase" << endl;
  }
  out_file << endl;
  out_file << "     end //always" << endl;
  out_file << "//synopsys translate_on" << endl;
  out_file << endl;
  out_file << get_checker_warning();
  out_file << "endmodule" << endl;
}

void Verilog_Printer::assigns_from_eqn(const string& eqn_file_name, const string& uniquify_string,
                                       vector<string>& inputs, vector<string>& outputs,
                                       vector<string>& assigns, vector<string>& wires) {
  ifstream in_file(eqn_file_name.c_str());
  if(!in_file.is_open()) {
    mout << "Failed to open input file: " << eqn_file_name << endl;
    return;
  }
  string line;
  vector<string> words;
  bool done = false;
  while(!done) {
    //mout << "line: " << line << endl;
    getline(in_file,line);
    if(line.find(";") != string::npos)
      done = true;
    line = line.substr(0,line.rfind(";"));
    splitString(line," \t=",words);
    for(unsigned int i = 1; i < words.size(); ++i)
      inputs.push_back(pointy_to_square_braces(words[i]));
  }
  done = false;
  while(!done) {
    //mout << "line: " << line << endl;
    getline(in_file,line);
    if(line.find(";") != string::npos)
      done = true;
    line = line.substr(0,line.rfind(";"));
    splitString(line," \t=",words);
    for(unsigned int i = 1; i < words.size(); ++i)
      outputs.push_back(pointy_to_square_braces(words[i]));
  }
        
  string cur_assign;
  string full_line;
  while(getline(in_file,line) && line.find("Don't care:") == string::npos) {
    //mout << "line: " << line << endl;
    if(line == "")
      continue;
    full_line = line;
    while(line.find(";") == string::npos && getline(in_file,line))
      full_line += line;
    full_line = full_line.substr(0,full_line.rfind(";"));
    replace_braces(full_line,uniquify_string);
    replace_nots(full_line);
    replace_ands(full_line);
    replace_ors(full_line);
    splitString(full_line,"=",words);
    string result = words[0];
    result = result.substr(result.find_first_not_of(" "),
                           result.find_last_not_of(" ") + 1);
    if(result.find(uniquify_string) == string::npos)
      result = uniquify_string + "_" + result;
    //          mout << "result: " << result << endl;
    string expr = words[1];
    //          mout << "expr: " << expr << endl;
    splitString(expr,"|",words);
    for(unsigned int i = 0; i < words.size(); ++i) {
      words[i] = words[i].substr(words[i].find_first_not_of(" "),
                                 words[i].find_last_not_of(" ") + 1);
      if(words[i] == "0")
        words[i] = "1'b0";
      else if(words[i] == "1")
        words[i] == "1'b1";
    }
    if(words.size() == 1)
      expr = words[0];
    else 
      expr = "(" + words[0] + ")";
    for(unsigned int i = 1; i < words.size(); ++i)
      expr += "\n\t\t\t| (" + words[i] + ")";
    cur_assign = result + " = " + expr;
    assigns.push_back(pointy_to_square_braces(cur_assign));
    //          mout << "cur_assign: " << cur_assign << endl;
    splitString(cur_assign," ",words);
    //          mout << "words[0]: " << words[0] << endl;
    wires.push_back(pointy_to_square_braces(result));
  }
}

