/**************************************************************************
 *   Inferno : An automatic semantic information extractor                 *
 *   Copyright (C) 2006-2010  University of Michigan                       *
 *   http://www.eecs.umich.edu/inferno                                     *
 *                                                                         *
 *   Contributors include Andrew DeOrio, Adam Bauserman,                   *
 *   Valeria Bertacco and Beth Isaksen.                                    *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA         *
 *   02110-1301  USA                                                       *
 *                                                                         *
 *   Andrew DeOrio <awdeorio@umich.edu>                                    *
 *   Valeria Bertacco <valeria@umich.edu>                                  *
 *                                                                         *
 *   University of Michigan                                                *
 *   Electrical Engineering and Computer Science Dept.                     *
 *   2260 Hayward St.                                                      *
 *   Ann Arbor, MI 48109-2121                                              *
 **************************************************************************/

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <map>
#include <set>
#include <time.h>
#include <limits.h>

#include "global_defs.h"
#include "util.h"
#include "System.h"
#include "global_vars.h"

using namespace std;

// global stream to allow output to cout and logfile at the same time
Multi_Stream mout;
// stream to write to logfile (also to screen, if verbose mode enabled)
Multi_Stream fout;
// file output stream
ofstream fileout;

int main(int argc, char* argv[])
{

  // check commandline arguments
  if(argc != 2) {
    cout << "Usage: " << argv[0] << " <inferno config file>" << endl;
    exit (1);
  }

  int start_time = clock(), cur_time, last_time = clock();
  data_t data, multi_bit_data, constants;
  dup_t duplicates;
        
  // if true, all log file output will also be sent to stdout
  bool verbose_mode = false;

  // set up output stream to write to screen and logfile
  mout.AddStream(&cout);
  string log_filename = "inferno.log";
  fileout.open(log_filename.c_str());
  mout.AddStream(&fileout);
  mout.SetMode(COUT | LOG);
        
  // set up other stream to write only to logfile (also to screen,
  // if in verbose mode)
  fout.AddStream(&cout);
  fout.AddStream(&fileout);
        
  if(verbose_mode) fout.SetMode(COUT | LOG);
  else fout.SetMode(LOG);
    

  mout << "********************************************************************************\n"
       << "*                                                                              *\n"
       << "*                                    Inferno                                   *\n"
       << "*                                                                              *\n"
       << "* by Andrew DeOrio, Adam Bauserman, Valeria Bertacco and Beth Isaksen          *\n"
       << "* http://www.eecs.umich.edu/inferno                                            *\n"
       << "* Copyright (C) University of Michigan                                         *\n"
       << "*                                                                              *\n"
       << "********************************************************************************\n"
       << endl;

  //check dependencies
  fout << "Checking for sis..." << endl;
        
  bool found_sis = find_in_path("sis");
  bool found_abc = find_in_path("abc");
        
  if (found_sis)
    fout << "Found sis." << endl;
  else if (found_abc)
    fout << "No sis, but found abc" << endl;
  else {
    mout << "Warning: can't find sis or abc.  Disabling Verilog checker generation." << endl;
  }

  // find out how big the numbers we can use are:
  fout << "Maximum unsigned long long int: " << ULLONG_MAX << endl;

  //We use a config file to read in the configuration for the run -- i.e. to tell
  //us what to analyze and how to analyze it
  char* config_file_name = argv[1];

  //Setup system
  System my_system(config_file_name);
  my_system.found_sis = found_sis;
  my_system.found_abc = found_abc;

  //Parse the config file (with some error checking)
  last_time = clock();
  if(!my_system.read_config_file()) {
    mout << "Error in config file." << endl;
    exit (1);
  }
        
  cur_time = clock();
  fout << "**Time to parse config file: " << (cur_time - last_time)/1000000 << "s\n";

  //Parse the verilog files -- we do this only to determine the inputs/outputs of the
  //modules and the connections among the modules -- in particular, we ignore the
  //actual logic (since we are not performing static analysis)
  last_time = clock();
  if(!my_system.parse_verilog()) {
    mout << "Error while parsing verilog." << endl;
    exit (1);
  }
  cur_time = clock();
  fout << "**Time to parse verilog: " << (cur_time - last_time)/1000000 << "s\n";

  //Parse the beginning of the trace file to determine the
  //translations from signal names to symbols
  last_time = clock();
  if(!my_system.parse_beginning_of_trace_file()) {
    mout << "Error while parsing beginning of trace file." << endl;
    exit (1);
  }
  cur_time = clock();
  fout << "**Time to parse beginning of trace file: " << (cur_time - last_time)/1000000 << "s\n";

  //Read the rest of the trace file in chunks (whose size is determined by max_lines)
  //to avoid filling up the memory of the machine
  //After each chunk, perform preliminary analysis -- i.e. find and record the "states"
  //and sequence of "states"
  last_time = clock();
        
  //    mout << "done parsing? " << my_system.done_parsing << endl;
        
  mout << "Reading values from VCD\n";
        
  string cmd;
  string dot_filename = my_system.output_file_base + string("transactions.dot");
                
  bool found_new;
        
  while(!my_system.done_parsing) {
    if(!my_system.parse_more_of_trace_file()) {
      mout << "Error while parsing rest of trace file." << endl;
      exit (1);
    }
    // mout << "done parsing? " << my_system.done_parsing << endl;

    if (INCREMENTAL_MODE) {
      //cout << "Finding transactions..." << endl;
      my_system.find_transactions();
      found_new = my_system.print_dot_file();
    }
  }
        
  cout << endl;
        
  cur_time = clock();
  mout << "**Time to parse rest of trace file: " << (cur_time - last_time)/1000000 << "s\n";

  //Once the entire trace has been read, perform the final analysis,
  //i.e. find the transactions, constants, legal transitions, etc. and print
  //results as dot files, verilog files, etc.
  cout << "Finding transactions..." << endl;
  my_system.find_transactions();
  my_system.find_constants();
        
  // find correlations of individual bits (experimental)
  my_system.find_correlations();
        
  cout << "Printing .dot files and verilog..." << endl;
  my_system.print_state_diagram();
  my_system.print_dot_file("", true);
        
  // convert DOT files to PDF
  string files[2] = {"transactions", "state"};
  for (int i = 0; i < 2; i++) { 
    string dot_file = my_system.output_file_base + files[i] + ".dot";
    string ps_file = my_system.output_file_base + files[i] + ".ps";
    string pdf_file = my_system.output_file_base + files[i] + ".pdf";
           
    //first, we oonvert DOT file to PS file
    cmd = string("dot -Tps -o " + ps_file + " " + dot_file);
    system(cmd.c_str());
           
    //next, we oonvert PS file into PDF file
    cmd = string("ps2pdf " + ps_file + " " + pdf_file);
    system(cmd.c_str());
           
    //we remove the PS file created
    cmd = string("rm " + ps_file);
    system(cmd.c_str());
  }
        
        
  if (found_sis || found_abc)
    my_system.print_verilog();

  fout << "System:\n" << my_system << endl;
  fout << "Done printing system.\n";

  mout << "Cleaning up files..." << endl;
  my_system.clean_files();

  my_system.print_summary();

  // copy log file and input file output dir
  cmd  = string("cp ") + log_filename + string(" ") + 
    my_system.output_file_base;
  system(cmd.c_str());
        
  cmd = "cp " + string(config_file_name) + string(" ") + 
    my_system.output_file_base; 
  system(cmd.c_str());

  mout << endl;
  mout << "to view output graphs, see pdf files in " << my_system.output_file_base << endl;

  mout << "**Total time: " << (cur_time - start_time)/1000000 << "s\n";

  // close log file
  fileout.close();
        
  return 0;
}
